"""
SequenceEncoder to convert sequence terms to sequence number.
"""

# Authors: Hongwei Zhang
# License: MIT


from collections import Counter
import numpy as np

from .base import SequenceColumn


class SequenceEncoder(SequenceColumn):
    """Encoder for sequence type feature. Convert terms to numbers.
    First index is 1.

    Parameters
    ----------
    sep : str, default=' '
        Separator of input sequence.

    min_cnt : int, default=5
        Minimum count of term.

    max_len: int, default=None
        Maximum length of sequence. If none is given,
        the maximum length of training sequence will be used.

    word2idx : dict
        Mappings from term to index.

    idx2word : dict
        Mappings from index to term.

    Attributes
    ----------
    sep : str, default=' '
        Separator of input sequence.

    min_cnt : int, default=5
        Minimum count of term.

    max_len: int, default=None
        Maximum length of sequence. If none is given,
        the maximum length of training sequence will be used.

    word2idx : dict
        Mappings from term to index.

    idx2word : dict
        Mappings from index to term.
    """
    def __init__(self, sep=' ', min_cnt=5, max_len=None,
                 word2idx=None, idx2word=None):
        self.sep = sep
        self.min_cnt = min_cnt
        self.max_len = max_len

        self.word2idx = word2idx if word2idx else dict()
        self.idx2word = idx2word if idx2word else dict()

    def fit(self, x, y=None):
        """Fit this transformer.

        Parameters
        ----------
        x : array-like
            One column of training data.
        y : array-like, default=None
            Training targets.

        Returns
        -------
        self : SequenceEncoder
            This SequenceEncoder.
        """

        if not self.word2idx:
            counter = Counter()

            max_len = 0
            for sequence in np.array(x).ravel():
                words = sequence.split(self.sep)
                counter.update(words)
                max_len = max(max_len, len(words))

            if self.max_len is None:
                self.max_len = max_len

            # drop rare words
            words = sorted(
                list(filter(lambda x: counter[x] >= self.min_cnt, counter)))

            self.word2idx = dict(zip(words, range(1, len(words) + 1)))
            self.word2idx['__PAD__'] = 0
            if '__UNKNOWN__' not in self.word2idx:
                self.word2idx['__UNKNOWN__'] = len(self.word2idx)

        if not self.idx2word:
            self.idx2word = {
                index: word for word, index in self.word2idx.items()}

        if not self.max_len:
            max_len = 0
            for sequence in np.array(x).ravel():
                words = sequence.split(self.sep)
                max_len = max(max_len, len(words))
            self.max_len = max_len

        return self

    def transform(self, x):
        """Transform x by this fitted transformer.

        Parameters
        ----------
        x : array-like
            Column data to be transformed.

        Returns
        -------
        transformed_x : array-like
            Transformed data.
        """
        transformed_x = list()

        for sequence in np.asarray(x).ravel():
            words = list()
            for word in sequence.split(self.sep):
                try:
                    words.append(self.word2idx[word])
                except KeyError:
                    words.append(self.word2idx['__UNKNOWN__'])

            transformed_x.append(
                np.asarray(words[0:self.max_len], dtype=np.int))

        return np.asarray(transformed_x, dtype=np.object)

    def dimension(self):
        """Number of unique terms.
        """
        return len(self.word2idx)

    def max_length(self):
        """Maximum length of one sequence.
        """
        return self.max_len



"""Dataset for torch.
"""

# Authors: Hongwei Zhang
# License: MIT


from collections import OrderedDict
from itertools import chain

import numpy as np
import torch.utils.data as data


class Dataset(data.Dataset):
    """Dataset for torch.

    Parameters
    ----------
    features : Features
        Fitted Features object.

    X_map : dict
        example:
            {'feature1': numpy.array([...]), 'feature2': numpy.array([...])}

    y : numpy.array
    """
    def __init__(self, features, X_map, y=None):
        self.features = features
        self.X_map = X_map
        self.y = y
        if y is not None:
            self.y = np.asarray(y, np.float32).reshape(-1, 1)

        self.__data_size = self.__get_data_size()

    def __get_data_size(self):
        key = next(iter(self.X_map))
        return self.X_map[key].shape[0]

    def __len__(self):
        return self.__data_size

    @staticmethod
    def __pad_sequence(sequence_feature, sequence):
        # zero is special index for padding
        padded_seq = np.zeros(sequence_feature.max_length(), np.int)
        padded_seq[0: sequence.shape[0]] = sequence

        return padded_seq

    def __getitem__(self, idx):
        record = OrderedDict()

        for feat in chain(
                self.features.number_features,
                self.features.category_features):
            record[feat.name] = self.X_map[feat.name][idx]

        for feat in self.features.sequence_features:
            seq = self.X_map[feat.name][idx]
            record[feat.name] = Dataset.__pad_sequence(feat, seq)
            record[f"__{feat.name}_length"] = seq.shape[0]

        if self.y is not None:
            record['label'] = self.y[idx]
        return record

    def get_num_batches(self, batch_size):
        return np.ceil(self.__data_size / batch_size)



from prediction_flow.pytorch.nn import MLP


def test_simple_creation():
    mlp = MLP(input_size=10, hidden_layers=(16, 4),
              activation=None, dropout=0.0)

    assert len(mlp._sequential) == 4


def test_creation_with_dropout():
    mlp = MLP(input_size=10, hidden_layers=(16, 4),
              activation=None, dropout=0.1)

    assert len(mlp._sequential) == 6


def test_creation_with_activation_and_dropout():
    mlp = MLP(input_size=10, hidden_layers=(16, 4),
              activation='relu', dropout=0.1)

    assert len(mlp._sequential) == 8



from prediction_flow.pytorch.nn import (
    AttentionGRUCell, AttentionUpdateGateGRUCell, DynamicGRU)

import torch
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence


def test_attention_gru_cell():
    gru_cell = AttentionGRUCell(10, 20)
    input = torch.randn(6, 3, 10)
    hx = torch.randn(3, 20)
    att_scores = torch.tensor([
        [0.1, 0.3, 0.6],
        [0.2, 0.2, 0.6],
        [0.1, 0.6, 0.3],
        [1.0, 0., 0.],
        [0.2, 0.3, 0.5],
        [0.1, 0.3, 0.6],
    ])

    output = []
    for i in range(6):
        hx = gru_cell(input[i], hx, att_scores[i])
        output.append(hx)

    assert len(output) == 6


def test_attention_update_gate_gru_cell():
    gru_cell = AttentionUpdateGateGRUCell(10, 20)
    input = torch.randn(6, 3, 10)
    hx = torch.randn(3, 20)
    att_scores = torch.tensor([
        [0.1, 0.3, 0.6],
        [0.2, 0.2, 0.6],
        [0.1, 0.6, 0.3],
        [1.0, 0., 0.],
        [0.2, 0.3, 0.5],
        [0.1, 0.3, 0.6],
    ])

    output = []
    for i in range(6):
        hx = gru_cell(input[i], hx, att_scores[i])
        output.append(hx)

    assert len(output) == 6


def test_dynamic_gru():
    rnn = DynamicGRU(3, 5)

    keys = torch.tensor([
        [[0.1, 0.2, 0.3], [1, 2, 3], [0.4, 0.2, 1], [0.0, 0.0, 0.0]],
        [[0.1, 0.2, 0.3], [1, 2, 3], [0.4, 0.2, 1], [0.5, 0.5, 0.5]],
        [[0.1, 0.2, 0.3], [0., 0., 0.], [0., 0., 0.], [0., 0., 0.]],
        [[0.1, 0.2, 0.3], [0., 0., 0.], [0., 0., 0.], [0., 0., 0.]],
        [[0.1, 0.2, 0.3], [0., 0., 0.], [0., 0., 0.], [0., 0., 0.]]
    ], dtype=torch.float)

    att_scores = torch.tensor([
        [0.0330, 0.7252, 0.2459, 0.],
        [0.2952, 0.8721, 0.4468, 0.0904],
        [0.4598,  0.,  0.,  0.],
        [0.0286, 0.,  0.,  0.],
        [0.0561, 0.,  0.,  0.]])

    lengths = torch.tensor([3, 4, 1, 1, 1])

    packed_att_scores = pack_padded_sequence(
        att_scores,
        lengths,
        batch_first=True, enforce_sorted=False)

    packed_keys = pack_padded_sequence(
        keys,
        lengths,
        batch_first=True, enforce_sorted=False)

    actual, actual_lengths = pad_packed_sequence(
        rnn(packed_keys, packed_att_scores), batch_first=True)

    assert actual.size() == (5, 4, 5)
    assert actual_lengths.numpy().tolist() == [3, 4, 1, 1, 1]



from prediction_flow.pytorch.nn import Attention


import numpy as np
import torch
import torch.nn.init as init


def test_attention():
    attention = Attention(3, [8], batchnorm=False, activation=0.0)

    query = torch.tensor([[1, 1, 1], [0.1, 0.2, 0.3]], dtype=torch.float)

    keys = torch.tensor([
        [[0.1, 0.2, 0.3], [1, 2, 3], [0.4, 0.2, 1], [0.0, 0.0, 0.0]],
        [[0.1, 0.2, 0.3], [1, 2, 3], [0.4, 0.2, 1], [0.5, 0.5, 0.5]]
    ], dtype=torch.float)

    keys_length = torch.tensor([3, 4])

    for param in attention.mlp.parameters():
        init.constant_(param, 1)

    for param in attention.fc.parameters():
        init.constant_(param, 1)

    output = attention(query, keys, keys_length)

    actual = output.detach().numpy()
    assert output.size()[0] == 2
    assert output.size()[1] == 3
    np.testing.assert_array_almost_equal(
        actual, np.array([[1.0, 2.0, 3.0],
                          [0.989024, 1.969694, 2.959199]], dtype=np.float))

def test_attention_return_scores():
    attention = Attention(
        3, [8], batchnorm=False, activation=0.0, return_scores=True)

    query = torch.tensor([[1, 1, 1], [0.1, 0.2, 0.3]], dtype=torch.float)

    keys = torch.tensor([
        [[0.1, 0.2, 0.3], [1, 2, 3], [0.4, 0.2, 1], [0.0, 0.0, 0.0]],
        [[0.1, 0.2, 0.3], [1, 2, 3], [0.4, 0.2, 1], [0.5, 0.5, 0.5]]
    ], dtype=torch.float)

    keys_length = torch.tensor([3, 4])

    for param in attention.mlp.parameters():
        init.constant_(param, 1)

    for param in attention.fc.parameters():
        init.constant_(param, 1)

    output = attention(query, keys, keys_length)

    actual = output.detach().numpy()

    expected = np.array(
        [[1.472415e-11, 1.000000e+00, 1.492623e-09, 0.000000e+00],
         [2.915521e-03, 9.821462e-01, 8.833572e-03, 6.104673e-03]],
        dtype=np.float)

    assert output.size()[0] == 2
    assert output.size()[1] == 4
    np.testing.assert_array_almost_equal(actual, expected)



"""
FM layer.
"""

# Authors: Hongwei Zhang
# License: MIT


import torch
import torch.nn as nn


class FM(nn.Module):
    """FM layer.
    """
    def __init__(self):
        super(FM, self).__init__()

    def forward(self, x):
        sum_squared = torch.pow(torch.sum(x, dim=1), 2).unsqueeze(1)
        squared_sum = torch.sum(torch.pow(x, 2), dim=1).unsqueeze(1)
        second_order = 0.5 * (sum_squared - squared_sum)
        return second_order



"""
Deep Interest Evolution Network.
"""

from collections import OrderedDict

import torch

from .nn import Interest
from .interest_net import InterestNet


class DIEN(InterestNet):
    """Deep Interest Evolution Network.

    Parameters
    ----------
    features : Features

    attention_groups : list of AttentionGroup

    num_classes : int
        Number of classes.

    embedding_size : int
        Size of embedding.

    hidden_layers : list
        Size of hidden layers.
        Example: [96, 32]

    dnn_activation : str
        Activation function of deep layers.
        Example: relu

    final_activation : str
        Activation function of output.

    dropout : float
        Dropout rate.

    use_negsampling : bool
    """
    def __init__(self, *args, use_negsampling=False, **kwargs):
        self.use_negsampling = use_negsampling
        super(DIEN, self).__init__(*args, **kwargs)

    def create_attention_fn(self, attention_group):
        return Interest(
            attention_group.pairs_count * self.embedding_size,
            gru_type=attention_group.gru_type,
            gru_dropout=attention_group.gru_dropout,
            att_hidden_layers=attention_group.hidden_layers,
            att_dropout=attention_group.att_dropout,
            att_activation=attention_group.activation,
            use_negsampling=self.use_negsampling)

    def forward(self, x):
        final_layer_inputs = list()

        # linear
        number_inputs = list()
        for feature in self.features.number_features:
            number_inputs.append(x[feature.name].view(-1, 1))

        embeddings = OrderedDict()
        for feature in self.features.category_features:
            embeddings[feature.name] = self.embeddings[
                feature.name](x[feature.name])

        for feature in self.features.sequence_features:
            if not self._is_attention_feature(feature):
                embeddings[feature.name] = self._sequence_poolings[
                    feature.name](self.embeddings[
                        feature.name](x[feature.name]))

        auxiliary_losses = []
        for attention_group in self.attention_groups:
            query = torch.cat(
                [embeddings[pair['ad']]
                 for pair in attention_group.pairs],
                dim=-1)
            pos_hist = torch.cat(
                [self.embeddings[pair['pos_hist']](
                    x[pair['pos_hist']]) for pair in attention_group.pairs],
                dim=-1)
            keys_length = torch.min(torch.cat(
                [x[f"__{pair['pos_hist']}_length"].view(-1, 1)
                 for pair in attention_group.pairs],
                dim=-1), dim=-1)[0]
            neg_hist = None
            if self.use_negsampling:
                neg_hist = torch.cat(
                    [self.embeddings[pair['neg_hist']](
                        x[pair['neg_hist']])
                     for pair in attention_group.pairs],
                    dim=-1)
            embeddings[attention_group.name], tmp_loss = (
                self._attention_poolings[attention_group.name](
                    query, pos_hist, keys_length, neg_hist))
            if tmp_loss is not None:
                auxiliary_losses.append(tmp_loss)

        emb_concat = torch.cat(number_inputs + [
            emb for emb in embeddings.values()], dim=-1)

        final_layer_inputs = self.mlp(emb_concat)

        output = self.final_layer(final_layer_inputs)

        if self.num_classes == 2 and self.final_activation == 'sigmoid':
            output = torch.sigmoid(output)
        elif self.num_classes > 1 and self.final_activation == 'softmax':
            output = torch.softmax(output)
        elif self.final_activation:
            raise NotImplementedError(
                f"pair (final_activation: {self.final_activation}, "
                f"num_classes: {self.num_classes}) is not implemented")

        auxiliary_avg_loss = None
        if auxiliary_losses:
            auxiliary_avg_loss = auxiliary_losses[0]
            size = len(auxiliary_losses)
            for i in range(1, size):
                auxiliary_avg_loss += auxiliary_losses[i]
            auxiliary_avg_loss /= size

        return output, auxiliary_avg_loss



"""
Iinterest part used by DIEN model.

Reference:
    Deep Interest Evolution Network for Click-Through Rate Prediction
    https://arxiv.org/pdf/1809.03672.pdf
"""

# Authors: Hongwei Zhang
# License: MIT


from collections import OrderedDict

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

from .attention import Attention
from .rnn import DynamicGRU


class AuxiliaryNet(nn.Module):
    """ NN for Auxiliary Loss.

    Parameters
    ----------
    input_size : int
        Size of input.

    hidden_layers : iterable
        Hidden layer sizes.

    activation : str
        Name of activation function. ReLU, PReLU and Sigmoid are supported.
    """
    def __init__(self, input_size, hidden_layers, activation='sigmoid'):
        super(AuxiliaryNet, self).__init__()
        modules = OrderedDict()

        previous_size = input_size
        for index, hidden_layer in enumerate(hidden_layers):
            modules[f"dense{index}"] = nn.Linear(previous_size, hidden_layer)
            if activation:
                if activation.lower() == 'relu':
                    modules[f"activation{index}"] = nn.ReLU()
                elif activation.lower() == 'prelu':
                    modules[f"activation{index}"] = nn.PReLU()
                elif activation.lower() == 'sigmoid':
                    modules[f"activation{index}"] = nn.Sigmoid()
                else:
                    raise NotImplementedError(f"{activation} is not supported")
            previous_size = hidden_layer
        modules["final_layer"] = nn.Linear(previous_size, 1)
        self._sequential = nn.Sequential(modules)

    def forward(self, input):
        return torch.sigmoid(self._sequential(input))


class Interest(nn.Module):
    """Interest layer.

    Parameters
    ----------
    input_size : int
        Size of input.

    gru_type : str
        Type of GRU. GRU, AIGRU, AGRU and AUGRU are supported.

    gru_dropout : float
        Dropout rate of GRU.

    att_hidden_layers : iterable
        Hidden layer sizes of attention.

    att_dropout : float
        Dropout rate of attention.

    att_batchnorm : bool
        Batchnorm of attention.

    att_activation : str
        Activation function name of attention.
        relu, prelu and sigmoid are supported.

    use_negsampling : bool
    """
    __SUPPORTED_GRU_TYPE__ = ['GRU', 'AIGRU', 'AGRU', 'AUGRU']

    def __init__(
            self,
            input_size,
            gru_type='GRU',
            gru_dropout=0.0,
            att_hidden_layers=[80, 40],
            att_dropout=0.0,
            att_batchnorm=True,
            att_activation='prelu',
            use_negsampling=False):
        super(Interest, self).__init__()
        if gru_type not in Interest.__SUPPORTED_GRU_TYPE__:
            raise NotImplementedError(f"gru_type: {gru_type} is not supported")

        self.gru_type = gru_type
        self.use_negsampling = use_negsampling

        self.interest_extractor = nn.GRU(
            input_size=input_size,
            hidden_size=input_size,
            batch_first=True,
            bidirectional=False)

        if self.use_negsampling:
            self.auxiliary_net = AuxiliaryNet(
                input_size * 2, hidden_layers=[100, 50])

        if gru_type == 'GRU':
            self.interest_evolution = nn.GRU(
                input_size=input_size,
                hidden_size=input_size,
                batch_first=True,
                bidirectional=False)

            self.attention = Attention(
                input_size=input_size,
                hidden_layers=att_hidden_layers,
                dropout=att_dropout,
                batchnorm=att_batchnorm,
                activation=att_activation)
        elif gru_type == 'AIGRU':
            self.attention = Attention(
                input_size=input_size,
                hidden_layers=att_hidden_layers,
                dropout=att_dropout,
                batchnorm=att_batchnorm,
                activation=att_activation,
                return_scores=True)

            self.interest_evolution = nn.GRU(
                input_size=input_size,
                hidden_size=input_size,
                batch_first=True,
                bidirectional=False)
        elif gru_type == 'AGRU' or gru_type == 'AUGRU':
            self.attention = Attention(
                input_size=input_size,
                hidden_layers=att_hidden_layers,
                dropout=att_dropout,
                batchnorm=att_batchnorm,
                activation=att_activation,
                return_scores=True)

            self.interest_evolution = DynamicGRU(
                input_size=input_size,
                hidden_size=input_size,
                gru_type=gru_type)

    @staticmethod
    def _get_last_state(states, keys_length):
        # states [B, T, H]
        batch_size, max_seq_length, hidden_size = states.size()

        mask = (torch.arange(max_seq_length, device=keys_length.device).repeat(
            batch_size, 1) == (keys_length.view(-1, 1) - 1))

        return states[mask]

    def cal_auxiliary_loss(
            self, states, click_seq, noclick_seq, keys_length):
        # states [B, T, H]
        # click_seq [B, T, H]
        # noclick_seq [B, T, H]
        # keys_length [B]
        batch_size, max_seq_length, embedding_size = states.size()

        mask = (torch.arange(max_seq_length, device=states.device).repeat(
            batch_size, 1) < keys_length.view(-1, 1)).float()

        click_input = torch.cat([states, click_seq], dim=-1)
        noclick_input = torch.cat([states, noclick_seq], dim=-1)
        embedding_size = embedding_size * 2

        click_p = self.auxiliary_net(
            click_input.view(
                batch_size * max_seq_length, embedding_size)).view(
                    batch_size, max_seq_length)[mask > 0].view(-1, 1)
        click_target = torch.ones(
            click_p.size(), dtype=torch.float, device=click_p.device)

        noclick_p = self.auxiliary_net(
            noclick_input.view(
                batch_size * max_seq_length, embedding_size)).view(
                    batch_size, max_seq_length)[mask > 0].view(-1, 1)
        noclick_target = torch.zeros(
            noclick_p.size(), dtype=torch.float, device=noclick_p.device)

        loss = F.binary_cross_entropy(
            torch.cat([click_p, noclick_p], dim=0),
            torch.cat([click_target, noclick_target], dim=0))

        return loss

    def forward(self, query, keys, keys_length, neg_keys=None):
        """
        Parameters
        ----------
        query: 2D tensor, [B, H]
        kerys: 3D tensor, [B, T, H]
        keys_length: 1D tensor, [B]
        neg_keys: 3D tensor, [B, T, H]

        Returns
        -------
        outputs: 2D tensor, [B, H]
        """
        batch_size, max_length, dim = keys.size()

        packed_keys = pack_padded_sequence(
            keys,
            lengths=keys_length.squeeze(),
            batch_first=True,
            enforce_sorted=False)

        packed_interests, _ = self.interest_extractor(packed_keys)

        aloss = None
        if (self.gru_type != 'GRU') or self.use_negsampling:
            interests, _ = pad_packed_sequence(
                packed_interests,
                batch_first=True,
                padding_value=0.0,
                total_length=max_length)

            if self.use_negsampling:
                aloss = self.cal_auxiliary_loss(
                    interests[:, :-1, :],
                    keys[:, 1:, :],
                    neg_keys[:, 1:, :],
                    keys_length - 1)

        if self.gru_type == 'GRU':
            packed_interests, _ = self.interest_evolution(packed_interests)

            interests, _ = pad_packed_sequence(
                packed_interests,
                batch_first=True,
                padding_value=0.0,
                total_length=max_length)

            outputs = self.attention(query, interests, keys_length)

        elif self.gru_type == 'AIGRU':
            # attention
            scores = self.attention(query, interests, keys_length)
            interests = interests * scores.unsqueeze(-1)

            packed_interests = pack_padded_sequence(
                interests,
                lengths=keys_length.squeeze(),
                batch_first=True,
                enforce_sorted=False)
            _, outputs = self.interest_evolution(packed_interests)
            outputs = outputs.squeeze()

        elif self.gru_type == 'AGRU' or self.gru_type == 'AUGRU':
            # attention
            scores = self.attention(query, interests, keys_length)

            packed_interests = pack_padded_sequence(
                interests,
                lengths=keys_length.squeeze(),
                batch_first=True,
                enforce_sorted=False)

            packed_scores = pack_padded_sequence(
                scores,
                lengths=keys_length.squeeze(),
                batch_first=True,
                enforce_sorted=False)

            outputs, _ = pad_packed_sequence(
                self.interest_evolution(
                    packed_interests, packed_scores), batch_first=True)
            # pick last state
            outputs = Interest._get_last_state(
                outputs, keys_length.squeeze())

        return outputs, aloss



"""
LogTransformer to convert number feature.
"""

# Authors: Hongwei Zhang
# License: MIT


import numpy as np

from .base import NumberColumn


class LogTransformer(NumberColumn):
    """LogTransformer to convert number feature.
    """
    def fit(self, x, y=None):
        """Fit this transformer.

        Parameters
        ----------
        x : array-like
            One column of training data.
        y : array-like, default=None, ignored
            Training targets.

        Returns
        -------
        self : LogTransformer
            This LogTransformer.
        """
        return self

    def transform(self, x):
        """ log(1 + x) when x > 0 else x

        Parameters
        ----------
        x : array-like
            Column data to be transformed.

        Returns
        ----------
        res: array-like
        """
        res = x.copy().astype(np.float).ravel()
        mask = x > 0.0
        res[mask] = np.log(1 + x[mask])
        return res



from .mlp import MLP
from .pooling import MaxPooling, SumPooling
from .fm import FM
from .attention import Attention
from .rnn import AttentionGRUCell, AttentionUpdateGateGRUCell, DynamicGRU
from .interest import Interest


__all__ = [
    'MLP',
    'MaxPooling',
    'SumPooling',
    'FM',
    'Attention',
    'AttentionGRUCell',
    'AttentionUpdateGateGRUCell',
    'DynamicGRU',
    'Interest'
]



from prediction_flow.pytorch.nn import FM

import numpy as np
import torch


def test_fm():
    fm = FM()

    x = torch.as_tensor([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
    actual = fm(x)

    # 11.0 = 1 * 2 + 1 * 3 + 2 * 3
    # 77.0 = 4 * 5 + 4 * 6 + 5 * 6
    np.testing.assert_array_almost_equal(
        actual.numpy(), np.array([[11.0], [74.0]], dtype=np.float))



from .deepfm import DeepFM
from .dnn import DNN
from .interest_net import AttentionGroup
from .din import DIN
from .dien import DIEN
from .wide_deep import WideDeep


__all__ = [
    'DeepFM',
    'DNN',
    'AttentionGroup',
    'DIN',
    'DIEN',
    'WideDeep'
]



from prediction_flow.transformers.column import StandardScaler

import numpy as np


def test_normal():
    scaler = StandardScaler()

    x = np.array([3, 4, 2, 24, 2], dtype=np.float)

    scaler.fit(x)

    actual = scaler.transform(x)
    expected = np.array([
        -0.46880723, -0.35160542, -0.58600904,  1.99243073, -0.58600904])

    np.testing.assert_array_almost_equal(actual, expected)



"""
Class to manage all features.
"""

# Authors: Hongwei Zhang
# License: MIT


from collections import OrderedDict


class Features(object):
    """Class to manage all features.

    Parameters
    ----------
    number_features : array-like
        NumberFeature array.

    category_features : array-like
        CategoryFeature array.

    sequence_features : array-like
        SequenceFeature array.

    Attributes
    ----------
    number_features : array-like
        NumberFeature array.

    category_features : array-like
        CategoryFeature array.

    sequence_features : array-like
        SequenceFeature array.

    """
    def __init__(
            self,
            number_features=[],
            category_features=[],
            sequence_features=[]):
        self.number_features = number_features
        self.category_features = category_features
        self.sequence_features = sequence_features

    def fit(self, df):
        """Fit all transformers.

        Parameters
        ----------
        df : pandas.DataFrame

        Returns
        ----------
        self : Features
        """
        for feature in (
                self.number_features +
                self.category_features +
                self.sequence_features):
            if feature.column_flow:
                feature.column_flow.fit(df[feature.name].values)

        return self

    def transform(self, df):
        """Transform df using fitted transformers.

        Parameters
        ----------
        df : pandas.DataFrame

        Returns
        ----------
        transformed_X: dict

        {'feature1': numpy.array([...]), 'feature2': numpy.array([...])}
        """
        transformed_X = OrderedDict()

        for feature in (
                self.number_features +
                self.category_features +
                self.sequence_features):
            if feature.column_flow:
                transformed_X[feature.name] = feature.column_flow.transform(
                    df[feature.name].values)
            else:
                transformed_X[feature.name] = df[feature.name].values

        return transformed_X

    def number_feature_names(self):
        return [feature.name for feature in self.number_features]

    def category_feature_names(self):
        return [feature.name for feature in self.category_features]

    def sequence_feature_names(self):
        return [feature.name for feature in self.sequence_features]



from prediction_flow.features import Number, Category, Sequence, Features
from prediction_flow.transformers.column import (
    StandardScaler, CategoryEncoder, SequenceEncoder)
from prediction_flow.pytorch.data import Dataset

import pandas as pd


__SAMPLE_DF = pd.DataFrame({
    'userId': [11, 11, 11, 11, 11],
    'userAge': [23, 21, 19, 17, 41],
    'movieId': [4226, 5971, 6291, 7153, 30707],
    'rating': [3.0, 2.0, 4.0, 4.6, 5.0],
    'timestamp': [1294796159, 1294796201, 1294796113, 1294796132, 1294796176],
    'title': ['Memento (2000)',
              'My Neighbor Totoro (Tonari no Totoro) (1988)',
              'Lilya 4-Ever (Lilja 4-ever) (2002)',
              'Lord of the Rings: The Return of the King, The (2003)',
              'Million Dollar Baby (2004)'],
    'genres': [
        'Mystery|Thriller',
        'Animation|Children|Drama|Fantasy',
        'Crime|Drama',
        'Action|Adventure|Drama|Fantasy',
        'Drama'],
    'topGenre': [
        'Mystery',
        'Animation',
        'Crime',
        'Action',
        'Drama'],
    'clickedMovieIds': [
        '5971|6291',
        '3242|42',
        '32|43542|3222|3',
        '',
        '34|23'],
    'clickedMovieTopGenres': [
        'Animation|Mystery',
        'Drama',
        'Drama',
        '',
        'Mystery|Crime'],
    'label': [1, 0, 0, 1, 0]})


def test_normal():
    number_features = [
        Number('userAge', StandardScaler()),
        Number('rating', StandardScaler())]

    category_features = [
        Category('userId', CategoryEncoder(min_cnt=1)),
        Category('movieId', CategoryEncoder(min_cnt=1)),
        Category('topGenre', CategoryEncoder(min_cnt=1))]

    sequence_features = [
        Sequence('title', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('genres', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('clickedMovieIds',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5)),
        Sequence('clickedMovieTopGenres',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5))]

    features = Features(
        number_features=number_features,
        category_features=category_features,
        sequence_features=sequence_features)

    features.fit(__SAMPLE_DF)

    X_map = features.transform(__SAMPLE_DF)

    dataset = Dataset(features, X_map, __SAMPLE_DF.label.values)

    assert dataset[0]['userId'] == 1
    assert dataset[0]['movieId'] == 1
    assert dataset[0]['genres'].tolist() == [8, 9, 0, 0]
    assert dataset[0]['__genres_length'] == 2
    assert dataset[0]['label'] == 1



from prediction_flow.features import Number, Category, Sequence, Features
from prediction_flow.transformers.column import (
    StandardScaler, CategoryEncoder, SequenceEncoder, ColumnFlow)

import numpy as np
import pandas as pd


__TEST_DATA = pd.DataFrame(
    [
        [23, 173, 'male', 'fish vegetable', 1],
        [43, 181, 'male', 'fish pork', 0],
        [35, 161, 'female', 'beef vegetable', 0],
        [41, 171, 'male', 'fish vegetable', 1],
        [16, 153, 'female', 'pork chicken vegetable', 0],
        [32, 168, 'female', 'fish beef', 1],
        [26, 177, 'male', 'chicken vegetable', 0],
        [76, 190, 'male', 'fish pork vegetable', 0]
    ],
    columns=['age', 'height', 'gender', 'likes', 'label'])


def test_simple_column_transformer_define():
    number_features = [
        Number('age', None),
        Number('height', StandardScaler())]

    category_features = [
        Category('gender', CategoryEncoder(min_cnt=1))]

    sequence_features = [
        Sequence('likes', SequenceEncoder(sep=' ', min_cnt=1))]

    features = Features(
        number_features, category_features, sequence_features)

    features.fit(__TEST_DATA)

    actual = features.transform(__TEST_DATA)

    expected_age = np.array([23, 43, 35, 41, 16, 32, 26, 76])
    expected_gender = np.array([2, 2, 1, 2, 1, 1, 2, 2])
    expected_height = np.array(
        [0.1159659, 0.85814767, -0.99730676, -0.06957954, -1.73948853,
         -0.34789771, 0.48705679, 1.69310217])

    assert len(actual) == 4
    assert features.number_feature_names() == ['age', 'height']
    assert features.category_feature_names() == ['gender']
    assert features.sequence_feature_names() == ['likes']
    np.testing.assert_array_equal(actual['age'], expected_age)
    np.testing.assert_array_equal(actual['gender'], expected_gender)
    np.testing.assert_array_almost_equal(actual['height'], expected_height)


def test_column_flow_define():
    number_features = [
        Number('age', None),
        Number('height', ColumnFlow([StandardScaler()]))]

    category_features = [
        Category('gender', ColumnFlow([CategoryEncoder(min_cnt=1)]))
    ]

    sequence_features = [
        Sequence('likes', ColumnFlow([SequenceEncoder(sep=' ', min_cnt=1)]))
    ]

    features = Features(
        number_features, category_features, sequence_features)

    features.fit(__TEST_DATA)

    actual = features.transform(__TEST_DATA)

    expected_age = np.array([23, 43, 35, 41, 16, 32, 26, 76])
    expected_gender = np.array([2, 2, 1, 2, 1, 1, 2, 2])
    expected_height = np.array(
        [0.1159659, 0.85814767, -0.99730676, -0.06957954, -1.73948853,
         -0.34789771, 0.48705679, 1.69310217])

    assert len(actual) == 4
    assert features.number_feature_names() == ['age', 'height']
    assert features.category_feature_names() == ['gender']
    assert features.sequence_feature_names() == ['likes']
    np.testing.assert_array_equal(actual['age'], expected_age)
    np.testing.assert_array_equal(actual['gender'], expected_gender)
    np.testing.assert_array_almost_equal(actual['height'], expected_height)



"""ColumnFlow contaions a chain of column-orientation
transformers (implementing fit/transform).
"""

# Authors: Hongwei Zhang
# License: MIT


import numpy as np


class ColumnFlow(object):
    """ColumnFlow contaions a chain of column-orientation
    transformers (implementing fit/transform).

    Parameters
    ----------
    transformers : list
        List of column transformers (implementing fit/transform) that are
        chained, in the order in which they are chained.

    verbose : boolean, optional
        If True, the log while fitting each transformer will be printed.

    Attributes
    ----------
    transformers : list
        List of column transformers (implementing fit/transform) that are
        chained, in the order in which they are chained.

    verbose : boolean, optional
        If True, the log while fitting each transformer will be printed.
    """

    def __init__(self, transformers, verbose=False):
        ColumnFlow.__check_transformers(transformers)
        self.transformers = transformers
        self.verbose = verbose

    @staticmethod
    def __check_transformers(transformers):
        if not isinstance(transformers, list):
            raise TypeError(
                "transformers must be list type, not {type(transformers)}")

        types = [
            transformer.column_type for transformer in transformers]

        if len(set(types)) != 1:
            raise ValueError("transformers must be the same type, not {types}")

    def fit(self, x, y=None):
        """Fit all transformers one after the other.

        Parameters
        ----------
        x : array-like
            One column of training data.
        y : array-like, default=None
            Training targets.

        Returns
        -------
        self : ColumnFlow
            This flow.
        """
        transformed_x = np.asarray(x).ravel()
        for transformer in self.transformers:
            transformer.fit(transformed_x, y)
            transformed_x = transformer.transform(transformed_x)

        return self

    def transform(self, x):
        """Transform x by all fitted transformers.

        Parameters
        ----------
        x : array-like
            Column data to be transformed.

        Returns
        -------
        transformed_x : array-like
            Transformed data.
        """
        transformed_x = np.asarray(x).ravel()
        for transformer in self.transformers:
            transformed_x = transformer.transform(transformed_x)

        return transformed_x



"""
Base class for features.
"""

# Authors: Hongwei Zhang
# License: MIT


from abc import ABC, abstractmethod
from ..transformers.column import Column, ColumnFlow


class BaseFeature(ABC):
    """Base class for all feature classes.

    Parameters
    ----------
    name : str
        Name of this feature.

    column_flow : ColumnFlow
        ColumnFlow to transform this feature.

    Attributes
    ----------
    name : str
        Name of this feature.

    column_flow : ColumnFlow or
                  list of column transformers or
                  single transformer
        ColumnFlow to transform this feature.
    """

    @abstractmethod
    def __init__(self, name, column_flow=None):
        self.name = name

        self.column_flow = None

        if column_flow:
            if isinstance(column_flow, ColumnFlow):
                self.column_flow = column_flow
            elif isinstance(column_flow, list):
                self.column_flow = ColumnFlow(column_flow)
            elif isinstance(column_flow, Column):
                self.column_flow = ColumnFlow([column_flow])
            else:
                raise NotImplementedError(
                    "column_flow should be ColumnFlow or "
                    "list of column transformers or single transformer")



import pytest

from prediction_flow.transformers.column import (
    LogTransformer, CategoryEncoder, ColumnFlow)


def test_wrong_type_transformers():
    with pytest.raises(TypeError):
        ColumnFlow({CategoryEncoder()})


def test_multi_type_transformers():
    with pytest.raises(ValueError):
        ColumnFlow([LogTransformer(), CategoryEncoder()])


def test_transformers():
    column_flow = ColumnFlow([CategoryEncoder(min_cnt=1)])

    input_terms = ['this', 'is', 'a', 'simple', 'test']

    column_flow.fit(input_terms)

    transformed = column_flow.transform(input_terms)

    assert set(transformed) == {1, 2, 3, 4, 5}
    assert column_flow.transformers[-1].dimension() == 7
    assert isinstance(input_terms, list) == True



"""
DeepFM.
"""
from collections import OrderedDict

import torch
import torch.nn as nn

from .base import EmbeddingMixin
from .nn import FM, MLP, MaxPooling
from .utils import init_weights


class DeepFM(nn.Module, EmbeddingMixin):
    """DeepFM.

    Parameters
    ----------
    features : Features

    num_classes : int
        Number of classes.

    embedding_size : int
        Size of embedding.

    hidden_layers : list
        Size of hidden layers.
        Example: [96, 32]

    activation : str
        Activation function.
        Example: relu

    final_activation : str
        Activation function of output.

    dropout : float
        Dropout rate.

    use_linear : bool

    use_fm : bool

    use_deep : bool
    """
    def __init__(self, features, num_classes, embedding_size, hidden_layers,
                 activation='relu', final_activation=None, dropout=None,
                 use_linear=True, use_fm=True, use_deep=True):
        super(DeepFM, self).__init__()
        self.features = features
        self.num_classes = num_classes
        self.final_activation = final_activation
        self.use_linear = use_linear
        self.use_fm = use_fm
        self.use_deep = use_deep

        self.embeddings, self.embedding_sizes = self.build_embeddings(
            embedding_size, fixed_embedding_size=True)

        self._sequence_poolings = OrderedDict()

        total_embedding_sizes = 0
        for feature in self.features.number_features:
            self.embeddings[feature.name] = nn.Linear(
                in_features=1,
                out_features=embedding_size,
                bias=False)
            self.add_module(
                f"embedding:{feature.name}",
                self.embeddings[feature.name])
            total_embedding_sizes += embedding_size

        for feature in self.features.category_features:
            total_embedding_sizes += (
                self.embedding_sizes[feature.name])

        for feature in self.features.sequence_features:
            self._sequence_poolings[feature.name] = MaxPooling(1)
            self.add_module(
                f"pooling:{feature.name}",
                self._sequence_poolings[feature.name])
            total_embedding_sizes += (
                self.embedding_sizes[feature.name])

        final_layer_input_size = 0
        # linear
        # This part is diff from deepFM paper,
        # sparse features are not included.
        self.linear = None
        if self.use_linear and self.features.number_features:
            self.linear = MLP(
                len(self.features.number_features),
                hidden_layers=[1], dropout=None, activation=None)
            final_layer_input_size += 1

        # fm
        self.fm = None
        if use_fm and total_embedding_sizes:
            self.fm = FM()
            final_layer_input_size += 1

        # deep
        self.mlp = None
        if use_deep and total_embedding_sizes:
            total_input_size = (total_embedding_sizes +
                                len(self.features.number_features))
            self.mlp = MLP(
                total_input_size, hidden_layers,
                dropout=dropout, activation=activation)
            final_layer_input_size += hidden_layers[-1]

        output_size = self.num_classes

        if self.num_classes == 2 and self.final_activation == 'sigmoid':
            output_size -= 1

        self.final_layer = nn.Linear(final_layer_input_size, output_size)

        self.apply(init_weights)

    def forward(self, x):
        final_layer_inputs = list()

        number_inputs = list()
        if self.linear:
            # linear
            for feature in self.features.number_features:
                number_inputs.append(x[feature.name].view(-1, 1))
            linear_concat = torch.cat(number_inputs, dim=1)
            final_layer_inputs.append(self.linear(linear_concat))

        embeddings = list()
        for feature in self.features.number_features:
            embeddings.append(
                self.embeddings[feature.name](
                    x[feature.name].view(-1, 1)))
        for feature in self.features.category_features:
            embeddings.append(
                self.embeddings[feature.name](x[feature.name]))
        for feature in self.features.sequence_features:
            embeddings.append(
                self._sequence_poolings[feature.name](
                    self.embeddings[feature.name](x[feature.name])))

        emb_concat = None
        if embeddings:
            emb_concat = torch.cat(embeddings, dim=1)

        # fm
        if self.fm:
            final_layer_inputs.append(self.fm(emb_concat))

        # deep
        if self.mlp:
            deep_input = torch.cat(number_inputs + [emb_concat], dim=1)
            final_layer_inputs.append(self.mlp(deep_input))

        final_layer_inputs = torch.cat(final_layer_inputs, dim=1)

        output = self.final_layer(final_layer_inputs)

        if self.num_classes == 2 and self.final_activation == 'sigmoid':
            output = torch.sigmoid(output)
        elif self.num_classes > 1 and self.final_activation == 'softmax':
            output = torch.softmax(output)
        elif self.final_activation:
            raise NotImplementedError(
                f"pair (final_activation: {self.final_activation}, "
                f"num_classes: {self.num_classes}) is not implemented")

        return output



from prediction_flow.pytorch.nn import Interest


import torch


def test_gru_interest_evolution():
    interests = Interest(
        input_size=3,
        gru_type='GRU',
        gru_dropout=0,
        att_hidden_layers=[8],
        att_dropout=0,
        att_batchnorm=False,
        att_activation=None)

    query = torch.tensor([[1, 1, 1], [0.1, 0.2, 0.3]], dtype=torch.float)

    keys = torch.tensor([
        [[0.1, 0.2, 0.3], [1, 2, 3], [0.4, 0.2, 1], [0.0, 0.0, 0.0]],
        [[0.1, 0.2, 0.3], [1, 2, 3], [0.4, 0.2, 1], [0.5, 0.5, 0.5]]
    ], dtype=torch.float)

    keys_length = torch.tensor([3, 4])

    output, _ = interests(query, keys, keys_length)

    assert output.size()[0] == 2
    assert output.size()[1] == 3


def test_aigru_interest_evolution():
    interests = Interest(
        input_size=3,
        gru_type='AIGRU',
        gru_dropout=0,
        att_hidden_layers=[8],
        att_dropout=0,
        att_batchnorm=False,
        att_activation=None)

    query = torch.tensor([[1, 1, 1], [0.1, 0.2, 0.3]], dtype=torch.float)

    keys = torch.tensor([
        [[0.1, 0.2, 0.3], [1, 2, 3], [0.4, 0.2, 1], [0.0, 0.0, 0.0]],
        [[0.1, 0.2, 0.3], [1, 2, 3], [0.4, 0.2, 1], [0.5, 0.5, 0.5]]
    ], dtype=torch.float)

    keys_length = torch.tensor([3, 4])

    output, _ = interests(query, keys, keys_length)

    assert output.size()[0] == 2
    assert output.size()[1] == 3


def test_agru_interest_evolution():
    interests = Interest(
        input_size=3,
        gru_type='AGRU',
        gru_dropout=0,
        att_hidden_layers=[8],
        att_dropout=0,
        att_batchnorm=False,
        att_activation=None)

    query = torch.tensor([[1, 1, 1], [0.1, 0.2, 0.3]], dtype=torch.float)

    keys = torch.tensor([
        [[0.1, 0.2, 0.3], [1, 2, 3], [0.4, 0.2, 1], [0.0, 0.0, 0.0]],
        [[0.1, 0.2, 0.3], [1, 2, 3], [0.4, 0.2, 1], [0.5, 0.5, 0.5]]
    ], dtype=torch.float)

    keys_length = torch.tensor([3, 4])

    output, _ = interests(query, keys, keys_length)

    assert output.size()[0] == 2
    assert output.size()[1] == 3


def test_augru_interest_evolution():
    interests = Interest(
        input_size=3,
        gru_type='AUGRU',
        gru_dropout=0,
        att_hidden_layers=[8],
        att_dropout=0,
        att_batchnorm=False,
        att_activation=None)

    query = torch.tensor([[1, 1, 1], [0.1, 0.2, 0.3]], dtype=torch.float)

    keys = torch.tensor([
        [[0.1, 0.2, 0.3], [1, 2, 3], [0.4, 0.2, 1], [0.0, 0.0, 0.0]],
        [[0.1, 0.2, 0.3], [1, 2, 3], [0.4, 0.2, 1], [0.5, 0.5, 0.5]]
    ], dtype=torch.float)

    keys_length = torch.tensor([3, 4])

    output, _ = interests(query, keys, keys_length)

    assert output.size()[0] == 2
    assert output.size()[1] == 3


def test_neg_sampling():
    interests = Interest(
        input_size=3,
        gru_type='AUGRU',
        gru_dropout=0,
        att_hidden_layers=[8],
        att_dropout=0,
        att_batchnorm=False,
        att_activation=None,
        use_negsampling=True)

    query = torch.tensor(
        [[1, 1, 1], [0.1, 0.2, 0.3], [0.3, 0.4, 0.5]], dtype=torch.float)

    keys = torch.tensor([
        [[0.1, 0.2, 0.3], [1, 2, 3], [0.4, 0.2, 1], [0.0, 0.0, 0.0]],
        [[0.1, 0.2, 0.3], [1, 2, 3], [0.4, 0.2, 1], [0.5, 0.5, 0.5]],
        [[0.1, 0.2, 0.3], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]
    ], dtype=torch.float)

    neg_keys = torch.tensor([
        [[0.3, 0.2, 0.1], [3, 2, 1], [1, 0.2, 0.4], [0.0, 0.0, 0.0]],
        [[0.3, 0.2, 0.1], [3, 2, 1], [1, 0.2, 0.4], [0.5, 0.5, 0.5]],
        [[0.3, 0.2, 0.1], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]
    ], dtype=torch.float)

    keys_length = torch.tensor([3, 4, 1])

    output, _ = interests(query, keys, keys_length, neg_keys)

    assert output.size()[0] == 3
    assert output.size()[1] == 3



"""
StandardScaler to convert term to number.
"""

# Authors: Hongwei Zhang
# License: MIT


import numpy as np
import sklearn.preprocessing as sk

from .base import NumberColumn


class StandardScaler(NumberColumn):
    """Normalize number feature.
    """
    def __init__(self):
        self.__scaler = sk.StandardScaler()

    def fit(self, x, y=None):
        """Fit this transformer.

        Parameters
        ----------
        x : array-like
            One column of training data.
        y : array-like, default=None, ignored
            Training targets.

        Returns
        -------
        self : StandardScaler
            This StandardScaler.
        """
        self.__scaler.fit(np.asarray(x, dtype=np.float).reshape(-1, 1))
        return self

    def transform(self, x):
        """Transform x by this fitted transformer.

        Parameters
        ----------
        x : array-like
            Column data to be transformed.

        Returns
        -------
        transformed_x : array-like
            Transformed data.
        """
        transformed_x = self.__scaler.transform(
            np.asarray(x, dtype=np.float32).reshape(-1, 1))

        return transformed_x.ravel()



from .base import Column
from .log_transformer import LogTransformer
from .standard_scaler import StandardScaler
from .category_encoder import CategoryEncoder
from .sequence_encoder import SequenceEncoder
from .column_flow import ColumnFlow


__all__ = [
    'StandardScaler',
    'LogTransformer',
    'CategoryEncoder',
    'SequenceEncoder',
    'ColumnFlow',
    'Column'
]



from prediction_flow.features import Number, Category, Sequence, Features
from prediction_flow.transformers.column import (
    StandardScaler, CategoryEncoder, SequenceEncoder)
from prediction_flow.pytorch import WideDeep


from .utils import prepare_dataloader


def test_normal():
    number_features = [
        Number('userAge', StandardScaler()),
        Number('rating', StandardScaler())]

    category_features = [
        Category('userId', CategoryEncoder(min_cnt=1)),
        Category('movieId', CategoryEncoder(min_cnt=1)),
        Category('topGenre', CategoryEncoder(min_cnt=1))]

    sequence_features = [
        Sequence('title', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('genres', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('clickedMovieIds',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5)),
        Sequence('clickedMovieTopGenres',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5))]

    features = Features(
        number_features=number_features,
        category_features=category_features,
        sequence_features=sequence_features)

    wide_features = ['rating', 'title', 'genres']
    deep_features = ['userAge', 'rating', 'userId', 'movieId', 'topGenre',
                     'clickedMovieIds', 'clickedMovieTopGenres']
    cross_features = [('movieId', 'clickedMovieIds'),
                      ('topGenre', 'clickedMovieTopGenres')]

    dataloader, _ = prepare_dataloader(features)

    model = WideDeep(
        features, wide_features, deep_features, cross_features,
        num_classes=2, embedding_size=4, hidden_layers=(8, 4),
        final_activation='sigmoid', dropout=0.3)

    model(next(iter(dataloader)))


def test_without_number_feature():
    number_features = []

    category_features = [
        Category('userId', CategoryEncoder(min_cnt=1)),
        Category('movieId', CategoryEncoder(min_cnt=1)),
        Category('topGenre', CategoryEncoder(min_cnt=1))]

    sequence_features = [
        Sequence('title', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('genres', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('clickedMovieIds',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5)),
        Sequence('clickedMovieTopGenres',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5))]

    features = Features(
        number_features=number_features,
        category_features=category_features,
        sequence_features=sequence_features)

    wide_features = ['title', 'genres']
    deep_features = ['userId', 'movieId', 'topGenre',
                     'clickedMovieIds', 'clickedMovieTopGenres']
    cross_features = [('movieId', 'clickedMovieIds'),
                      ('topGenre', 'clickedMovieTopGenres')]

    dataloader, _ = prepare_dataloader(features)

    model = WideDeep(
        features, wide_features, deep_features, cross_features,
        num_classes=2, embedding_size=4, hidden_layers=(8, 4),
        final_activation='sigmoid', dropout=0.3)

    model(next(iter(dataloader)))


def test_without_category_feature():
    number_features = []

    category_features = []

    sequence_features = [
        Sequence('title', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('genres', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('clickedMovieIds',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5)),
        Sequence('clickedMovieTopGenres',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5))]

    features = Features(
        number_features=number_features,
        category_features=category_features,
        sequence_features=sequence_features)

    wide_features = ['title', 'genres']
    deep_features = ['clickedMovieIds', 'clickedMovieTopGenres']

    dataloader, _ = prepare_dataloader(features)

    model = WideDeep(
        features, wide_features, deep_features, [],
        num_classes=2, embedding_size=4, hidden_layers=(8, 4),
        final_activation='sigmoid', dropout=0.3)

    model(next(iter(dataloader)))


def test_only_with_number_features():
    number_features = [
        Number('userAge', StandardScaler()),
        Number('rating', StandardScaler())]

    category_features = []

    sequence_features = []

    features = Features(
        number_features=number_features,
        category_features=category_features,
        sequence_features=sequence_features)

    wide_features = ['rating', 'userAge']

    dataloader, _ = prepare_dataloader(features)

    model = WideDeep(
        features, wide_features, [], [],
        num_classes=2, embedding_size=4, hidden_layers=(8, 4),
        final_activation='sigmoid', dropout=0.3)

    model(next(iter(dataloader)))



"""
CatagoryEncoder to convert term to number.
"""

# Authors: Hongwei Zhang
# License: MIT


import numpy as np
from collections import Counter

from .base import CategoryColumn


class CategoryEncoder(CategoryColumn):
    """Encoder for category type feature.

    Parameters
    ----------
    min_cnt : int, default=5
        Minimum count of term.

    word2idx : dict
        Mappings from term to index.

    idx2word : dict
        Mappings from index to term.

    Attributes
    ----------
    min_cnt : int, default=5
        Minimum count of term.

    word2idx : dict
        Mappings from term to index.

    idx2word : dict
        Mappings from index to term.
    """
    def __init__(self, min_cnt=5, word2idx=None, idx2word=None):
        self.min_cnt = min_cnt

        self.word2idx = word2idx if word2idx else dict()
        self.idx2word = idx2word if idx2word else dict()

    def fit(self, x, y=None):
        """Fit this transformer.

        Parameters
        ----------
        x : array-like
            One column of training data.
        y : array-like, default=None, ignored
            Training targets.

        Returns
        -------
        self : CategoryEncoder
            This CategoryEncoder.
        """
        if not self.word2idx:
            counter = Counter(np.asarray(x).ravel())

            selected_terms = sorted(
                list(filter(lambda x: counter[x] >= self.min_cnt, counter)))

            self.word2idx = dict(
                zip(selected_terms, range(1, len(selected_terms) + 1)))
            self.word2idx['__PAD__'] = 0
            if '__UNKNOWN__' not in self.word2idx:
                self.word2idx['__UNKNOWN__'] = len(self.word2idx)

        if not self.idx2word:
            self.idx2word = {
                index: word for word, index in self.word2idx.items()}

        return self

    def transform(self, x):
        """Transform x by this fitted transformer.

        Parameters
        ----------
        x : array-like
            Column data to be transformed.

        Returns
        -------
        transformed_x : array-like
            Transformed data.
        """
        transformed_x = list()
        for term in np.asarray(x).ravel():
            try:
                transformed_x.append(self.word2idx[term])
            except KeyError:
                transformed_x.append(self.word2idx['__UNKNOWN__'])

        return np.asarray(transformed_x, dtype=np.int)

    def dimension(self):
        return len(self.word2idx)



"""
Attention used by DIN model.

Reference:
    Deep Interest Network for Click-Through Rate Prediction
    https://github.com/zhougr1993/DeepInterestNetwork/blob/master/din/model.py
"""

# Authors: Hongwei Zhang
# License: MIT


import numpy as np

import torch
import torch.nn as nn
from torch.functional import F

from .mlp import MLP


class Attention(nn.Module):
    """Attention layer.

    Parameters
    ----------
    input_size : int
        Size of input.

    hidden_layers : iterable
        Hidden layer sizes.

    dropout : float
        Dropout rate.

    activation : str
        Name of activation function. relu, prelu and sigmoid are supported.

    return_scores : bool
        Return attention scores instead of weighted sum pooling result.
    """
    def __init__(
            self,
            input_size,
            hidden_layers,
            dropout=0.0,
            batchnorm=True,
            activation='prelu',
            return_scores=False):
        super(Attention, self).__init__()
        self.return_scores = return_scores
        self.mlp = MLP(
            input_size=input_size * 4,
            hidden_layers=hidden_layers,
            dropout=dropout,
            batchnorm=batchnorm,
            activation=activation)
        self.fc = nn.Linear(hidden_layers[-1], 1)

    def forward(self, query, keys, keys_length):
        """
        Parameters
        ----------
        query: 2D tensor, [B, H]
        kerys: 3D tensor, [B, T, H]
        keys_length: 1D tensor, [B]

        Returns
        -------
        outputs: 2D tensor, if return_scores=False [B, H], otherwise [B, T]
        """
        batch_size, max_length, dim = keys.size()

        query = query.unsqueeze(1).expand(-1, max_length, -1)

        din_all = torch.cat(
            [query, keys, query - keys, query * keys], dim=-1)

        din_all = din_all.view(batch_size * max_length, -1)

        outputs = self.mlp(din_all)

        outputs = self.fc(outputs).view(batch_size, max_length)  # [B, T]

        # Scale
        outputs = outputs / (dim ** 0.5)

        # Mask
        mask = (torch.arange(max_length, device=keys_length.device).repeat(
            batch_size, 1) < keys_length.view(-1, 1))
        outputs[~mask] = -np.inf

        # Activation
        outputs = F.softmax(outputs, dim=1)  # [B, T]

        if not self.return_scores:
            # Weighted sum
            outputs = torch.matmul(
                outputs.unsqueeze(1), keys).squeeze()  # [B, H]

        return outputs



"""
Base class for all column-orientation transformer classes
with fit/transform functions.
"""

# Authors: Hongwei Zhang
# License: MIT


from abc import ABC, abstractmethod
from enum import Enum


class Column(ABC):
    """Base class for all column-orientation transformer classes
    with fit/transform functions.
    """

    @abstractmethod
    def fit(self, x, y=None):
        """Fit this transformer.

        Parameters
        ----------
        x : array-like
            One column of training data.
        y : array-like, default=None
            Training targets.
        """

        raise NotImplementedError

    @abstractmethod
    def transform(self, x):
        """Transform x by this fitted transformer.

        Parameters
        ----------
        x : array-like
            Column data to be transformed.
        """

        raise NotImplementedError


class ColumnType(Enum):
    NUMBER = 1
    CATEGORY = 2
    SEQUENCE = 3


class NumberColumn(Column):
    """Base class for all column-orientation number type transformer classes
    with fit/transform functions.
    """
    column_type = ColumnType.NUMBER


class CategoryColumn(Column):
    """Base class for all column-orientation category type transformer classes
    with fit/transform functions.
    """
    column_type = ColumnType.CATEGORY

    @abstractmethod
    def dimension(self):
        """Number of unique terms.
        """
        raise NotImplementedError


class SequenceColumn(Column):
    """Base class for all column-orientation sequence type transformer classes
    with fit/transform functions.
    """
    column_type = ColumnType.SEQUENCE

    @abstractmethod
    def dimension(self):
        """Number of unique terms.
        """
        raise NotImplementedError

    @abstractmethod
    def max_length(self):
        """Maximum length of one sequence.
        """
        raise NotImplementedError



from prediction_flow.features import Number, Category, Sequence, Features
from prediction_flow.transformers.column import (
    StandardScaler, CategoryEncoder, SequenceEncoder)
from prediction_flow.pytorch import AttentionGroup, DIN


from .utils import prepare_dataloader


def test_normal():
    number_features = [
        Number('userAge', StandardScaler()),
        Number('rating', StandardScaler())]

    category_features = [
        Category('userId', CategoryEncoder(min_cnt=1)),
        Category('movieId', CategoryEncoder(min_cnt=1)),
        Category('topGenre', CategoryEncoder(min_cnt=1))]

    sequence_features = [
        Sequence('title', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('genres', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('clickedMovieIds',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5)),
        Sequence('clickedMovieTopGenres',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5))]

    attention_groups = [
        AttentionGroup(
            name='group1',
            pairs=[{'ad': 'movieId', 'pos_hist': 'clickedMovieIds'},
                   {'ad': 'topGenre', 'pos_hist': 'clickedMovieTopGenres'}],
            hidden_layers=[8, 4])]

    features = Features(
        number_features=number_features,
        category_features=category_features,
        sequence_features=sequence_features)

    dataloader, _ = prepare_dataloader(features)

    model = DIN(
        features, attention_groups=attention_groups,
        num_classes=2, embedding_size=4, hidden_layers=(16, 8),
        final_activation='sigmoid', dropout=0.3)

    model(next(iter(dataloader)))



import torch.nn as nn
import torch.nn.init as init


def init_weights(model):
    if isinstance(model, nn.Linear):
        if model.weight is not None:
            init.kaiming_uniform_(model.weight.data)
        if model.bias is not None:
            init.normal_(model.bias.data)
    elif isinstance(model, nn.BatchNorm1d):
        if model.weight is not None:
            init.normal_(model.weight.data, mean=1, std=0.02)
        if model.bias is not None:
            init.constant_(model.bias.data, 0)
    elif isinstance(model, nn.BatchNorm2d):
        if model.weight is not None:
            init.normal_(model.weight.data, mean=1, std=0.02)
        if model.bias is not None:
            init.constant_(model.bias.data, 0)
    elif isinstance(model, nn.BatchNorm3d):
        if model.weight is not None:
            init.normal_(model.weight.data, mean=1, std=0.02)
        if model.bias is not None:
            init.constant_(model.bias.data, 0)
    else:
        pass



import numpy as np

from prediction_flow.features import Number, Category, Sequence, Features
from prediction_flow.transformers.column import (
    StandardScaler, CategoryEncoder, SequenceEncoder)
from prediction_flow.pytorch import DNN


from .utils import prepare_dataloader, _SAMPLE_DF


def test_normal():
    number_features = [
        Number('userAge', StandardScaler()),
        Number('rating', StandardScaler())]

    category_features = [
        Category('userId', CategoryEncoder(min_cnt=1)),
        Category('movieId', CategoryEncoder(min_cnt=1)),
        Category('topGenre', CategoryEncoder(min_cnt=1))]

    sequence_features = [
        Sequence('title', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('genres', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('clickedMovieIds',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5)),
        Sequence('clickedMovieTopGenres',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5))]

    features = Features(
        number_features=number_features,
        category_features=category_features,
        sequence_features=sequence_features)

    dataloader, _ = prepare_dataloader(features)

    model = DNN(
        features, num_classes=2, embedding_size=4, hidden_layers=(8, 4),
        final_activation='sigmoid', dropout=0.3)

    model(next(iter(dataloader)))


def test_without_number_feature():
    number_features = []

    category_features = [
        Category('userId', CategoryEncoder(min_cnt=1)),
        Category('movieId', CategoryEncoder(min_cnt=1)),
        Category('topGenre', CategoryEncoder(min_cnt=1))]

    sequence_features = [
        Sequence('title', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('genres', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('clickedMovieIds',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5)),
        Sequence('clickedMovieTopGenres',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5))]

    features = Features(
        number_features=number_features,
        category_features=category_features,
        sequence_features=sequence_features)

    dataloader, _ = prepare_dataloader(features)

    model = DNN(
        features, num_classes=2, embedding_size=4, hidden_layers=(8, 4),
        final_activation='sigmoid', dropout=0.3)

    model(next(iter(dataloader)))


def test_without_category_feature():
    number_features = []

    category_features = []

    sequence_features = [
        Sequence('title', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('genres', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('clickedMovieIds',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5)),
        Sequence('clickedMovieTopGenres',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5))]

    features = Features(
        number_features=number_features,
        category_features=category_features,
        sequence_features=sequence_features)

    dataloader, _ = prepare_dataloader(features)

    model = DNN(
        features, num_classes=2, embedding_size=4, hidden_layers=(8, 4),
        final_activation='sigmoid', dropout=0.3)

    model(next(iter(dataloader)))


def test_only_with_number_features():
    number_features = [
        Number('userAge', StandardScaler()),
        Number('rating', StandardScaler())]

    category_features = []

    sequence_features = []

    features = Features(
        number_features=number_features,
        category_features=category_features,
        sequence_features=sequence_features)

    dataloader, _ = prepare_dataloader(features)

    model = DNN(
        features, num_classes=2, embedding_size=4, hidden_layers=(8, 4),
        final_activation='sigmoid', dropout=0.3)

    model(next(iter(dataloader)))


def test_shared_embedding():
    number_features = []

    movie_enc = SequenceEncoder(sep='|', min_cnt=1, max_len=5)
    genre_enc = SequenceEncoder(sep='|', min_cnt=1, max_len=5)

    movie_enc.fit(
        np.concatenate((_SAMPLE_DF.clickedMovieIds.values,
                        _SAMPLE_DF.movieId.values), axis=None))

    genre_enc.fit(
        np.concatenate((_SAMPLE_DF.clickedMovieTopGenres.values,
                        _SAMPLE_DF.topGenre.values), axis=None))

    category_features = [
        Category('userId', CategoryEncoder(min_cnt=1)),
        Category('movieId',
                 CategoryEncoder(
                     min_cnt=1,
                     word2idx=movie_enc.word2idx,
                     idx2word=movie_enc.idx2word),
                 embedding_name='movieId'),
        Category('topGenre',
                 CategoryEncoder(
                     min_cnt=1,
                     word2idx=genre_enc.word2idx,
                     idx2word=genre_enc.idx2word),
                 embedding_name='topGenre', embedding_size=8)]

    sequence_features = [
        Sequence('title', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('genres', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('clickedMovieIds',
                 SequenceEncoder(
                     sep='|',
                     min_cnt=1,
                     max_len=5,
                     word2idx=movie_enc.word2idx,
                     idx2word=movie_enc.idx2word),
                 embedding_name='movieId'),
        Sequence('clickedMovieTopGenres',
                 SequenceEncoder(
                     sep='|',
                     min_cnt=1,
                     max_len=5,
                     word2idx=genre_enc.word2idx,
                     idx2word=genre_enc.idx2word),
                 embedding_name='topGenre', embedding_size=8)]

    features = Features(
        number_features=number_features,
        category_features=category_features,
        sequence_features=sequence_features)

    dataloader, _ = prepare_dataloader(features)

    model = DNN(
        features, num_classes=2, embedding_size=16, hidden_layers=(8, 4),
        final_activation='sigmoid', dropout=0.3)

    model(next(iter(dataloader)))



"""AttentionGRU and AttentionUpdateGateGRU.
"""

# Authors: Hongwei Zhang
# License: MIT


import math

import torch
import torch.nn as nn
import torch.nn.init as init
import torch.nn.functional as F
from torch.nn.utils.rnn import PackedSequence


class AttentionGRUCell(nn.Module):
    def __init__(self, input_size, hidden_size, bias=True):
        super(AttentionGRUCell, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.bias = bias
        # (W_ir|W_iz|W_in)
        self.weight_ih = nn.Parameter(
            torch.Tensor(3 * hidden_size, input_size))
        # (W_hr|W_hz|W_hn)
        self.weight_hh = nn.Parameter(
            torch.Tensor(3 * hidden_size, hidden_size))
        if bias:
            # (b_ir|b_iz|b_in)
            self.bias_ih = nn.Parameter(torch.Tensor(3 * hidden_size))
            # (b_hr|b_hz|b_hn)
            self.bias_hh = nn.Parameter(torch.Tensor(3 * hidden_size))
        else:
            self.register_parameter('bias_ih', None)
            self.register_parameter('bias_hh', None)
        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.hidden_size)
        for weight in self.parameters():
            init.uniform_(weight, -stdv, stdv)

    def forward(self, input, hx, att_score):
        """

        References
        ----------
            https://github.com/pytorch/pytorch/blob/v0.4.1/torch/nn/_functions/rnn.py#L49
        """

        gi = F.linear(input, self.weight_ih, self.bias_ih)
        gh = F.linear(hx, self.weight_hh, self.bias_hh)
        i_r, i_z, i_n = gi.chunk(3, 1)
        h_r, h_z, h_n = gh.chunk(3, 1)

        resetgate = torch.sigmoid(i_r + h_r)
        # updategate = torch.sigmoid(i_z + h_z)
        newgate = torch.tanh(i_n + resetgate * h_n)
        # hy = newgate + updategate * (hx - newgate)

        att_score = att_score.view(-1, 1)

        hy = (1. - att_score) * hx + att_score * newgate

        return hy


class AttentionUpdateGateGRUCell(nn.Module):
    def __init__(self, input_size, hidden_size, bias=True):
        super(AttentionUpdateGateGRUCell, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.bias = bias
        # (W_ir|W_iz|W_in)
        self.weight_ih = nn.Parameter(
            torch.Tensor(3 * hidden_size, input_size))
        # (W_hr|W_hz|W_hn)
        self.weight_hh = nn.Parameter(
            torch.Tensor(3 * hidden_size, hidden_size))
        if bias:
            # (b_ir|b_iz|b_in)
            self.bias_ih = nn.Parameter(torch.Tensor(3 * hidden_size))
            # (b_hr|b_hz|b_hn)
            self.bias_hh = nn.Parameter(torch.Tensor(3 * hidden_size))
        else:
            self.register_parameter('bias_ih', None)
            self.register_parameter('bias_hh', None)
        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.hidden_size)
        for weight in self.parameters():
            init.uniform_(weight, -stdv, stdv)

    def forward(self, input, hx, att_score):
        """

        References
        ----------
            https://github.com/pytorch/pytorch/blob/v0.4.1/torch/nn/_functions/rnn.py#L49
        """

        gi = F.linear(input, self.weight_ih, self.bias_ih)
        gh = F.linear(hx, self.weight_hh, self.bias_hh)
        i_r, i_z, i_n = gi.chunk(3, 1)
        h_r, h_z, h_n = gh.chunk(3, 1)

        resetgate = torch.sigmoid(i_r + h_r)
        updategate = torch.sigmoid(i_z + h_z)
        newgate = torch.tanh(i_n + resetgate * h_n)

        updategate = att_score.view(-1, 1) * updategate

        hy = newgate + updategate * (hx - newgate)

        return hy


class DynamicGRU(nn.Module):
    def __init__(self, input_size, hidden_size, bias=True, gru_type='AGRU'):
        super(DynamicGRU, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size

        if gru_type == 'AGRU':
            self.rnn = AttentionGRUCell(input_size, hidden_size, bias)
        elif gru_type == 'AUGRU':
            self.rnn = AttentionUpdateGateGRUCell(
                input_size, hidden_size, bias)

    def forward(self, input, att_scores, hx=None):
        is_packed_input = isinstance(input, PackedSequence)
        if not is_packed_input:
            raise NotImplementedError(
                "DynamicGRU only supports packed input")

        is_packed_att_scores = isinstance(att_scores, PackedSequence)
        if not is_packed_att_scores:
            raise NotImplementedError(
                "DynamicGRU only supports packed att_scores")

        input, batch_sizes, sorted_indices, unsorted_indices = input
        att_scores, _, _, _ = att_scores

        max_batch_size = batch_sizes[0]
        max_batch_size = int(max_batch_size)

        if hx is None:
            hx = torch.zeros(
                max_batch_size, self.hidden_size,
                dtype=input.dtype, device=input.device)

        outputs = torch.zeros(
            input.size(0), self.hidden_size,
            dtype=input.dtype, device=input.device)

        begin = 0
        for batch in batch_sizes:
            new_hx = self.rnn(
                input[begin: begin + batch],
                hx[0:batch],
                att_scores[begin: begin + batch])
            outputs[begin: begin + batch] = new_hx
            hx = new_hx
            begin += batch

        return PackedSequence(
            outputs, batch_sizes, sorted_indices, unsorted_indices)



"""
Deep Interest Network.
"""

from .nn import Attention
from .interest_net import InterestNet


class DIN(InterestNet):
    """Deep Interest Network.

    Parameters
    ----------
    features : Features

    attention_groups : list of AttentionGroup

    num_classes : int
        Number of classes.

    embedding_size : int
        Size of embedding.

    hidden_layers : list
        Size of hidden layers.
        Example: [96, 32]

    dnn_activation : str
        Activation function of deep layers.
        Example: relu

    final_activation : str
        Activation function of output.

    dropout : float
        Dropout rate.
    """
    def __init__(self, *args, **kwargs):
        super(DIN, self).__init__(*args, **kwargs)

    def create_attention_fn(self, attention_group):
        return Attention(
            attention_group.pairs_count * self.embedding_size,
            hidden_layers=attention_group.hidden_layers,
            dropout=attention_group.att_dropout,
            activation=attention_group.activation)



"""
Class for number feature.
"""

# Authors: Hongwei Zhang
# License: MIT


from .base import BaseFeature


class Number(BaseFeature):
    """
    Class for number feature.

    Parameters
    ----------
    name : str
        Name of this feature.

    column_flow : ColumnFlow
        ColumnFlow to transform this feature.

    Attributes
    ----------
    name : str
        Name of this feature.

    column_flow : ColumnFlow
        ColumnFlow to transform this feature.
    """
    def __init__(self, name, column_flow):
        super().__init__(name=name, column_flow=column_flow)



"""
Wide&Deep Model.
"""
from collections import OrderedDict

import torch
import torch.nn as nn

from .base import EmbeddingMixin
from .nn import MLP, SumPooling
from .utils import init_weights


class WideDeep(nn.Module, EmbeddingMixin):
    """Wide&Deep Model.

    Parameters
    ----------
    features : Features

    wide_features : list of str
        Feature names for wide part.

    deep_features : list of str
        Feature names for deep part.

    cross_features: list of tuple
        Cross sparse feature names for wide part.

    num_classes : int
        Number of classes.

    embedding_size : int
        Size of embedding.

    hidden_layers : list
        Size of hidden layers.
        Example: [96, 32]

    activation : str
        Activation function.
        Example: relu

    final_activation : str
        Activation function of output.

    dropout : float
        Dropout rate.
    """
    def __init__(self, features, wide_features, deep_features, cross_features,
                 num_classes, embedding_size, hidden_layers,
                 activation='relu', final_activation=None, dropout=0.0):
        super(WideDeep, self).__init__()
        self.features = features
        self.wide_features = wide_features
        self.deep_features = deep_features
        self.cross_features = cross_features
        self.num_classes = num_classes
        self.final_activation = final_activation

        self.embeddings, self.embedding_sizes = self.build_embeddings(
            embedding_size)

        self._sequence_poolings = OrderedDict()

        wide_input_size = 0
        deep_input_size = 0

        for feature in self.features.number_features:
            if feature.name in self.wide_features:
                wide_input_size += 1
            if feature.name in self.deep_features:
                deep_input_size += 1

        for feature in self.features.category_features:
            if feature.name in self.wide_features:
                wide_input_size += self.embedding_sizes[feature.name]
            if feature.name in self.deep_features:
                deep_input_size += self.embedding_sizes[feature.name]

        for feature in self.features.sequence_features:
            self._sequence_poolings[feature.name] = SumPooling(1)
            self.add_module(
                f"pooling:{feature.name}",
                self._sequence_poolings[feature.name])
            if feature.name in self.wide_features:
                wide_input_size += self.embedding_sizes[feature.name]
            if feature.name in self.deep_features:
                deep_input_size += self.embedding_sizes[feature.name]

        # plus cross embedding size
        wide_input_size += len(self.cross_features) * embedding_size

        final_layer_input_size = wide_input_size

        if deep_input_size:
            self.mlp = MLP(
                deep_input_size,
                hidden_layers,
                dropout=dropout, batchnorm=True, activation=activation)
            final_layer_input_size += hidden_layers[-1]

        output_size = self.num_classes

        if self.num_classes == 2 and self.final_activation == 'sigmoid':
            output_size -= 1

        self.final_layer = nn.Linear(final_layer_input_size, output_size)

        self.apply(init_weights)

    def forward(self, x):
        wide_inputs = list()
        deep_inputs = list()
        cross_inputs = list()

        for feature in self.features.number_features:
            if feature.name in self.wide_features:
                wide_inputs.append(x[feature.name].view(-1, 1))
            if feature.name in self.deep_features:
                deep_inputs.append(x[feature.name].view(-1, 1))

        for feature in self.features.category_features:
            if feature.name in self.wide_features:
                wide_inputs.append(
                    self.embeddings[feature.name](x[feature.name]))
            if feature.name in self.deep_features:
                deep_inputs.append(
                    self.embeddings[feature.name](x[feature.name]))

        for feature in self.features.sequence_features:
            if feature.name in self.wide_features:
                wide_inputs.append(
                    self._sequence_poolings[feature.name](
                        self.embeddings[feature.name](
                            x[feature.name])))
            if feature.name in self.deep_features:
                deep_inputs.append(
                    self._sequence_poolings[feature.name](
                        self.embeddings[feature.name](
                            x[feature.name])))

        # prepare cross features
        for x_f, y_f in self.cross_features:
            if x_f in self._sequence_poolings:
                x_emb = self._sequence_poolings[x_f](
                    self.embeddings[x_f](x[x_f]))
            else:
                x_emb = self.embeddings[x_f](x[x_f])

            if y_f in self._sequence_poolings:
                y_emb = self._sequence_poolings[y_f](
                    self.embeddings[y_f](x[y_f]))
            else:
                y_emb = self.embeddings[y_f](x[y_f])
            cross_inputs.append(x_emb * y_emb)

        final_layer_inputs = list()
        if wide_inputs:
            final_layer_inputs.append(torch.cat(wide_inputs, dim=1))

        if cross_inputs:
            final_layer_inputs.append(torch.cat(cross_inputs, dim=1))

        if deep_inputs:
            final_layer_inputs.append(self.mlp(torch.cat(deep_inputs, dim=1)))

        output = self.final_layer(torch.cat(final_layer_inputs, dim=1))

        if self.num_classes == 2 and self.final_activation == 'sigmoid':
            output = torch.sigmoid(output)
        elif self.num_classes > 1 and self.final_activation == 'softmax':
            output = torch.softmax(output)
        elif self.final_activation:
            raise NotImplementedError(
                f"pair (final_activation: {self.final_activation}, "
                f"num_classes: {self.num_classes}) is not implemented")

        return output



from prediction_flow.transformers.column import SequenceEncoder

import numpy as np


def test_normal():
    sequence_encoder = SequenceEncoder(sep=' ', min_cnt=1, max_len=3)

    x = [
        "this is a simple test",
        "this class is work"
    ]

    sequence_encoder.fit(x)

    actual = sequence_encoder.transform(x)
    assert sequence_encoder.dimension() == 9
    assert sequence_encoder.max_length() == 3
    assert actual.tolist() == [[6, 3, 1], [6, 2, 3]]
    assert isinstance(x, list)


def test_unseen_inputs():
    sequence_encoder = SequenceEncoder(sep=' ', min_cnt=1, max_len=10)

    x = [
        "this is a simple test",
        "this class is work"
    ]

    sequence_encoder.fit(x)

    actual = sequence_encoder.transform(["this is an unseen test"])
    assert actual.tolist() == [[6, 3, 8, 8, 5]]



"""Interest Net.
"""

from collections import OrderedDict

import torch
import torch.nn as nn

from .base import EmbeddingMixin
from .nn import MLP, MaxPooling
from .utils import init_weights


class AttentionGroup(object):
    """ This class is used to identify which features should be
    processed by attention. All candidate features and all behavior
    sequential features must be the same embedding size. All behavior
    sequential features must be the same maximum length.

    Parameters
    ----------
    name : str
        Unique group name.

    pairs : list of dict
        Example :
            [{'ad': 'item_id',
              'pos_hist': 'clicked_item_ids',
              'neg_hist': 'neg_item_ids'},
             {'ad': 'item_category',
              'pos_hist': 'clicked_item_categories',
              'neg_hist': 'neg_item_categories'}]

    hidden_layers : iterable
        Hidden layer sizes of attention.

    activation : str
        Activation function of attention.
        Example: prelu

    att_dropout : float
        Dropout rate of attention.

    gru_type : str
        Type of GRU. GRU, AIGRU, AGRU and AUGRU are supported.

    gru_dropout : float
        Dropout rate of GRU.
    """
    def __init__(self, name, pairs,
                 hidden_layers, activation='prelu', att_dropout=0.0,
                 gru_type='GRU', gru_dropout=0.0):
        self.name = name
        self.pairs = pairs
        self.hidden_layers = hidden_layers
        self.activation = activation
        self.att_dropout = att_dropout
        self.gru_type = gru_type
        self.gru_dropout = gru_dropout

        self.related_feature_names = set()
        self.neg_feature_names = set()
        for pair in pairs:
            self.related_feature_names.add(pair['ad'])
            self.related_feature_names.add(pair['pos_hist'])
            if 'neg_hist' in pair:
                self.related_feature_names.add(pair['neg_hist'])
                self.neg_feature_names.add(pair['neg_hist'])

    def is_attention_feature(self, feature_name):
        if feature_name in self.related_feature_names:
            return True
        return False

    def is_neg_sampling_feature(self, feature_name):
        if feature_name in self.neg_feature_names:
            return True
        return False

    @property
    def pairs_count(self):
        return len(self.pairs)


class InterestNet(nn.Module, EmbeddingMixin):
    """Interest Network.

    Parameters
    ----------
    features : Features

    attention_groups : list of AttentionGroup

    num_classes : int
        Number of classes.

    embedding_size : int
        Size of embedding.

    hidden_layers : list
        Size of hidden layers.
        Example: [96, 32]

    dnn_activation : str
        Activation function of deep layers.
        Example: relu

    final_activation : str
        Activation function of output.

    dropout : float
        Dropout rate.
    """
    def _is_attention_feature(self, feature):
        for group in self.attention_groups:
            if group.is_attention_feature(feature.name):
                return True
        return False

    def _is_neg_sampling_feature(self, feature):
        for group in self.attention_groups:
            if group.is_neg_sampling_feature(feature.name):
                return True
        return False

    def create_attention_fn(self, attention_group):
        raise NotImplementedError(
            "Please implement the func to create attention")

    def __init__(self, features, attention_groups, num_classes, embedding_size,
                 hidden_layers, dnn_activation='prelu', final_activation=None,
                 dropout=0.0):
        super(InterestNet, self).__init__()
        self.features = features
        self.attention_groups = attention_groups
        self.num_classes = num_classes
        self.embedding_size = embedding_size
        self.hidden_layers = hidden_layers
        self.dnn_activation = dnn_activation
        self.final_activation = final_activation
        self.dropout = dropout

        self.embeddings, self.embedding_sizes = self.build_embeddings(
            embedding_size)

        self._sequence_poolings = OrderedDict()
        self._attention_poolings = OrderedDict()

        total_embedding_sizes = 0
        for feature in self.features.category_features:
            total_embedding_sizes += (
                self.embedding_sizes[feature.name])

        for feature in self.features.sequence_features:
            if not self._is_neg_sampling_feature(feature):
                total_embedding_sizes += (
                    self.embedding_sizes[feature.name])
            if not self._is_attention_feature(feature):
                self._sequence_poolings[feature.name] = MaxPooling(1)
                self.add_module(
                    f"pooling:{feature.name}",
                    self._sequence_poolings[feature.name])

        # attention
        for attention_group in self.attention_groups:
            self._attention_poolings[attention_group.name] = (
                self.create_attention_fn(attention_group))
            self.add_module(
                f"attention_pooling:{attention_group.name}",
                self._attention_poolings[attention_group.name])

        total_input_size = (total_embedding_sizes +
                            len(self.features.number_features))
        self.mlp = MLP(
            total_input_size,
            hidden_layers,
            dropout=dropout, batchnorm=True, activation=dnn_activation)
        final_layer_input_size = hidden_layers[-1]

        output_size = self.num_classes

        if self.num_classes == 2 and self.final_activation == 'sigmoid':
            output_size -= 1

        self.final_layer = nn.Linear(final_layer_input_size, output_size)

        self.apply(init_weights)

    def forward(self, x):
        final_layer_inputs = list()

        # linear
        number_inputs = list()
        for feature in self.features.number_features:
            number_inputs.append(x[feature.name].view(-1, 1))

        embeddings = OrderedDict()
        for feature in self.features.category_features:
            embeddings[feature.name] = self.embeddings[
                feature.name](x[feature.name])

        for feature in self.features.sequence_features:
            if not self._is_attention_feature(feature):
                embeddings[feature.name] = self._sequence_poolings[
                    feature.name](self.embeddings[
                        feature.name](x[feature.name]))

        for attention_group in self.attention_groups:
            query = torch.cat(
                [embeddings[pair['ad']]
                 for pair in attention_group.pairs],
                dim=-1)
            keys = torch.cat(
                [self.embeddings[pair['pos_hist']](
                    x[pair['pos_hist']]) for pair in attention_group.pairs],
                dim=-1)
            keys_length = torch.min(torch.cat(
                [x[f"__{pair['pos_hist']}_length"].view(-1, 1)
                 for pair in attention_group.pairs],
                dim=-1), dim=-1)[0]
            embeddings[attention_group.name] = self._attention_poolings[
                attention_group.name](query, keys, keys_length)

        emb_concat = torch.cat(number_inputs + [
            emb for emb in embeddings.values()], dim=-1)

        final_layer_inputs = self.mlp(emb_concat)

        output = self.final_layer(final_layer_inputs)

        if self.num_classes == 2 and self.final_activation == 'sigmoid':
            output = torch.sigmoid(output)
        elif self.num_classes > 1 and self.final_activation == 'softmax':
            output = torch.softmax(output)
        elif self.final_activation:
            raise NotImplementedError(
                f"pair (final_activation: {self.final_activation}, "
                f"num_classes: {self.num_classes}) is not implemented")

        return output



from prediction_flow.pytorch.data import Dataset


import pandas as pd
import torch.utils.data as data


_SAMPLE_DF = pd.DataFrame({
    'userId': [11, 11, 11, 11, 11],
    'userAge': [23, 21, 19, 17, 41],
    'movieId': ['4226', '5971', '6291', '7153', '30707'],
    'rating': [3.0, 2.0, 4.0, 4.6, 5.0],
    'timestamp': [1294796159, 1294796201, 1294796113, 1294796132, 1294796176],
    'title': ['Memento (2000)',
              'My Neighbor Totoro (Tonari no Totoro) (1988)',
              'Lilya 4-Ever (Lilja 4-ever) (2002)',
              'Lord of the Rings: The Return of the King, The (2003)',
              'Million Dollar Baby (2004)'],
    'genres': [
        'Mystery|Thriller',
        'Animation|Children|Drama|Fantasy',
        'Crime|Drama',
        'Action|Adventure|Drama|Fantasy',
        'Drama'],
    'topGenre': [
        'Mystery',
        'Animation',
        'Crime',
        'Action',
        'Drama'],
    'clickedMovieIds': [
        '5971|6291',
        '3242|42',
        '32|43542|3222|3',
        '',
        '34|23'],
    'clickedMovieTopGenres': [
        'Animation|Mystery',
        'Drama',
        'Drama|Drama|Drama|Drama',
        '',
        'Mystery|Crime'],
    'noClickedMovieIds': [
        '233|291',
        '324|421',
        '325|3542|322|33',
        '',
        '45|48'],
    'noClickedMovieTopGenres': [
        'Drama|Crime',
        'Animation|Mystery',
        'Mystery|Animation|Crime|Drama',
        '',
        'Crime|Mystery'],
    'label': [1, 0, 0, 1, 0]})


def prepare_dataloader(features):
    features.fit(_SAMPLE_DF)

    X_map = features.transform(_SAMPLE_DF)

    dataset = Dataset(features, X_map, _SAMPLE_DF.label.values)

    dataloader = data.DataLoader(
        dataset, batch_size=_SAMPLE_DF.shape[0], shuffle=False)

    return dataloader, _SAMPLE_DF



"""
Deep Neural Network.
"""
from collections import OrderedDict

import torch
import torch.nn as nn

from .base import EmbeddingMixin
from .nn import MLP, MaxPooling
from .utils import init_weights


class DNN(nn.Module, EmbeddingMixin):
    """Deep Neural Network.

    Parameters
    ----------
    features : Features

    num_classes : int
        Number of classes.

    embedding_size : int
        Size of embedding.

    hidden_layers : list
        Size of hidden layers.
        Example: [96, 32]

    activation : str
        Activation function.
        Example: relu

    final_activation : str
        Activation function of output.

    dropout : float
        Dropout rate.
    """
    def __init__(self, features, num_classes, embedding_size, hidden_layers,
                 activation='relu', final_activation=None, dropout=0.0):
        super(DNN, self).__init__()
        self.features = features
        self.num_classes = num_classes
        self.final_activation = final_activation

        self.embeddings, self.embedding_sizes = self.build_embeddings(
            embedding_size)

        self._sequence_poolings = OrderedDict()

        total_embedding_sizes = 0
        for feature in self.features.category_features:
            total_embedding_sizes += (
                self.embedding_sizes[feature.name])

        for feature in self.features.sequence_features:
            self._sequence_poolings[feature.name] = MaxPooling(1)
            self.add_module(
                f"pooling:{feature.name}",
                self._sequence_poolings[feature.name])
            total_embedding_sizes += (
                self.embedding_sizes[feature.name])

        total_input_size = (total_embedding_sizes +
                            len(self.features.number_features))
        self.mlp = MLP(
            total_input_size,
            hidden_layers,
            dropout=dropout, batchnorm=True, activation=activation)
        final_layer_input_size = hidden_layers[-1]

        output_size = self.num_classes

        if self.num_classes == 2 and self.final_activation == 'sigmoid':
            output_size -= 1

        self.final_layer = nn.Linear(final_layer_input_size, output_size)

        self.apply(init_weights)

    def forward(self, x):
        final_layer_inputs = list()

        # linear
        number_inputs = list()
        for feature in self.features.number_features:
            number_inputs.append(x[feature.name].view(-1, 1))

        embeddings = list()
        for feature in self.features.category_features:
            embeddings.append(
                self.embeddings[feature.name](x[feature.name]))

        for feature in self.features.sequence_features:
            embeddings.append(
                self._sequence_poolings[feature.name](
                    self.embeddings[feature.name](x[feature.name])))

        emb_concat = torch.cat(number_inputs + embeddings, dim=1)

        final_layer_inputs = self.mlp(emb_concat)

        output = self.final_layer(final_layer_inputs)

        if self.num_classes == 2 and self.final_activation == 'sigmoid':
            output = torch.sigmoid(output)
        elif self.num_classes > 1 and self.final_activation == 'softmax':
            output = torch.softmax(output)
        elif self.final_activation:
            raise NotImplementedError(
                f"pair (final_activation: {self.final_activation}, "
                f"num_classes: {self.num_classes}) is not implemented")

        return output



import numpy as np

from prediction_flow.transformers.column import LogTransformer


def test_normal():
    log_transformer = LogTransformer()

    x = np.array([100, 10, 32])
    log_transformer.fit(x)

    np.testing.assert_array_almost_equal(
        log_transformer.transform(x), np.array([4.615121, 2.397895, 3.496508]))



from tqdm import tqdm_notebook, tqdm

import numpy as np

import torch
import torch.utils.data as data

from prediction_flow.pytorch.data import Dataset
from prediction_flow.features import Features


def __to_gpu(device, batch):
    for key, tensor in batch.items():
        batch[key] = tensor.to(device)


def fit(epochs, model, loss, optimizer, train_loader,
        valid_loader=None, notebook=False,
        auxiliary_loss_rate=0.0):
    if notebook:
        epoch_bar = tqdm_notebook(
            desc='training routine', total=epochs, position=0)
        train_bar = tqdm_notebook(
            desc='train', total=len(train_loader), position=1)
        if valid_loader:
            valid_bar = tqdm_notebook(
                desc='valid', total=len(valid_loader), position=2)
    else:
        epoch_bar = tqdm(
            desc='training routine', total=epochs, position=0)
        train_bar = tqdm(
            desc='train', total=len(train_loader), position=1)
        if valid_loader:
            valid_bar = tqdm(
                desc='valid', total=len(valid_loader), position=2)

    use_cuda = torch.cuda.is_available()
    device = torch.device("cuda:0" if use_cuda else "cpu")

    if use_cuda:
        print("GPU is available, transfer model to GPU.")
        model = model.to(device)

    losses = []

    for epoch in range(epochs):
        model.train()
        running_loss = 0
        auxiliary_running_loss = 0
        for index, batch in enumerate(train_loader):
            if use_cuda:
                __to_gpu(device, batch)
            label = batch['label']
            # step 1. zero the gradients
            optimizer.zero_grad()
            # step 2. compute the output
            pred = model(batch)
            if isinstance(pred, tuple):
                pred, auxiliary_loss = pred
                if auxiliary_loss:
                    auxiliary_running_loss += (
                        (auxiliary_loss.item() -
                         auxiliary_running_loss) / (index + 1))
            # step 3. compute the loss
            loss_t = loss(pred, label)
            if isinstance(pred, tuple):
                if auxiliary_loss:
                    loss_t += auxiliary_loss_rate * auxiliary_loss
            running_loss += (loss_t.item() - running_loss) / (index + 1)
            # step 4. use loss to produce gradients
            loss_t.backward()
            # step 5. use optimizer to take gradient step
            optimizer.step()
            # update bar
            train_bar.set_postfix(loss=running_loss, epoch=index)
            train_bar.update()
        train_bar.reset()
        train_loss = running_loss
        train_auxiliary_loss = auxiliary_running_loss

        valid_loss = 0
        if valid_loader:
            model.eval()
            running_loss = 0
            auxiliary_running_loss = 0
            with torch.no_grad():
                for index, batch in enumerate(valid_loader):
                    if use_cuda:
                        __to_gpu(device, batch)
                    label = batch['label']
                    # step 1 compute the output
                    pred = model(batch)
                    # step 2. compute the loss
                    auxiliary_loss = torch.tensor(0.0)
                    pred = model(batch)
                    if isinstance(pred, tuple):
                        pred, auxiliary_loss = pred
                        if auxiliary_loss:
                            auxiliary_running_loss += (
                                (auxiliary_loss.item() -
                                 auxiliary_running_loss) / (index + 1))
                    loss_t = loss(pred, label)
                    if isinstance(pred, tuple):
                        if auxiliary_loss:
                            loss_t += auxiliary_loss_rate * auxiliary_loss
                    running_loss += (
                        loss_t.item() - running_loss) / (index + 1)
                    # update bar
                    valid_bar.set_postfix(
                        loss=running_loss, epoch=index)
                    valid_bar.update()
                valid_loss = running_loss
                valid_auxiliary_loss = auxiliary_running_loss
            valid_bar.reset()

        epoch_bar.set_postfix(
            train_loss=train_loss, valid_loss=valid_loss, epoch=epoch)
        epoch_bar.update()
        losses.append(
            {'train_loss': train_loss,
             'valid_loss': valid_loss,
             'train_auxiliary_loss': train_auxiliary_loss,
             'valid_auxiliary_loss': valid_auxiliary_loss})

    return losses


def predict(model, test_loader):
    use_cuda = torch.cuda.is_available()
    device = torch.device("cuda:0" if use_cuda else "cpu")

    model.zero_grad()
    model.eval()

    preds = list()
    with torch.no_grad():
        for _, batch in enumerate(test_loader):
            if use_cuda:
                __to_gpu(device, batch)
            # step 1 compute the output
            pred = model(batch)
            if isinstance(pred, tuple):
                pred, auxiliary_loss = pred
            preds.append(pred.cpu().numpy())

    return np.vstack(preds)


def create_dataloader_fn(
        number_features, category_features, sequence_features, batch_size,
        train_df, label_col='label', test_df=None, num_workers=0):

    features = Features(
        number_features=number_features,
        category_features=category_features,
        sequence_features=sequence_features)

    features = features.fit(train_df)

    train_X_map = features.transform(train_df)
    train_y = train_df[label_col].values
    train_dataset = Dataset(features, train_X_map, train_y)
    train_loader = data.DataLoader(
        train_dataset, batch_size=batch_size,
        shuffle=True, num_workers=num_workers)

    test_loader = None
    if test_df is not None:
        test_X_map = features.transform(test_df)
        test_y = None
        if label_col in set(test_df.columns):
            test_y = test_df[label_col].values
        test_dataset = Dataset(features, test_X_map, test_y)
        test_loader = data.DataLoader(
            test_dataset, batch_size=batch_size,
            shuffle=False, num_workers=num_workers)

    return features, train_loader, test_loader



from prediction_flow.features import Number, Category, Sequence, Features
from prediction_flow.transformers.column import (
    StandardScaler, CategoryEncoder, SequenceEncoder)
from prediction_flow.pytorch import AttentionGroup, DIEN


from .utils import prepare_dataloader


def create_test_data():
    number_features = [
        Number('userAge', StandardScaler()),
        Number('rating', StandardScaler())]

    category_features = [
        Category('userId', CategoryEncoder(min_cnt=1)),
        Category('movieId', CategoryEncoder(min_cnt=1)),
        Category('topGenre', CategoryEncoder(min_cnt=1))]

    sequence_features = [
        Sequence('title', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('genres', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('clickedMovieIds',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5)),
        Sequence('clickedMovieTopGenres',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5)),
        Sequence('noClickedMovieIds',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5)),
        Sequence('noClickedMovieTopGenres',
                 SequenceEncoder(sep='|', min_cnt=1, max_len=5))]

    attention_groups = [
        AttentionGroup(
            name='group1',
            pairs=[{'ad': 'movieId',
                    'pos_hist': 'clickedMovieIds',
                    'neg_hist': 'noClickedMovieIds'},
                   {'ad': 'topGenre',
                    'pos_hist': 'clickedMovieTopGenres',
                    'neg_hist': 'noClickedMovieTopGenres'}],
            hidden_layers=[8, 4])]

    features = Features(
        number_features=number_features,
        category_features=category_features,
        sequence_features=sequence_features)

    dataloader, _ = prepare_dataloader(features)

    return dataloader, features, attention_groups


def test_gru_gru_att():
    dataloader, features, attention_groups = create_test_data()

    attention_groups[0].gru_type = 'GRU'

    model = DIEN(
        features, attention_groups=attention_groups,
        num_classes=2, embedding_size=4, hidden_layers=(16, 8),
        final_activation='sigmoid', dropout=0.3)

    model(next(iter(dataloader)))


def test_gru_att_gru():
    dataloader, features, attention_groups = create_test_data()

    attention_groups[0].gru_type = 'AIGRU'

    model = DIEN(
        features, attention_groups=attention_groups,
        num_classes=2, embedding_size=4, hidden_layers=(16, 8),
        final_activation='sigmoid', dropout=0.3)

    model(next(iter(dataloader)))


def test_gru_agru():
    dataloader, features, attention_groups = create_test_data()

    attention_groups[0].gru_type = 'AGRU'

    model = DIEN(
        features, attention_groups=attention_groups,
        num_classes=2, embedding_size=4, hidden_layers=(16, 8),
        final_activation='sigmoid', dropout=0.3)

    model(next(iter(dataloader)))


def test_gru_augru():
    dataloader, features, attention_groups = create_test_data()

    attention_groups[0].gru_type = 'AUGRU'

    model = DIEN(
        features, attention_groups=attention_groups,
        num_classes=2, embedding_size=4, hidden_layers=(16, 8),
        final_activation='sigmoid', dropout=0.3)

    model(next(iter(dataloader)))


def test_gru_augru_neg():
    dataloader, features, attention_groups = create_test_data()

    attention_groups[0].gru_type = 'AUGRU'

    model = DIEN(
        features, attention_groups=attention_groups,
        use_negsampling=True,
        num_classes=2, embedding_size=4, hidden_layers=(16, 8),
        final_activation='sigmoid', dropout=0.3)

    model(next(iter(dataloader)))


def create_test_data_with_sharing_emb():
    number_features = [
        Number('userAge', StandardScaler()),
        Number('rating', StandardScaler())]

    # provide word to index mapping
    movie_word2idx = {
        '__PAD__': 0,
        '4226': 1,
        '5971': 2,
        '6291': 3,
        '7153': 4,
        '30707': 5,
        '3242': 6,
        '42': 7,
        '32': 8,
        '34': 9,
        '233': 10,
        '291': 11,
        '324': 12,
        '325': 13,
        '3542': 14,
        '322': 15,
        '33': 16,
        '45': 17,
        '__UNKNOWN__': 18}

    movie_idx2word = {
        index: word for word, index in movie_word2idx.items()}

    category_features = [
        Category('movieId',
                 CategoryEncoder(
                     word2idx=movie_word2idx,
                     idx2word=movie_idx2word),
                 embedding_name='movieId'),
        Category('topGenre', CategoryEncoder(min_cnt=1))]

    sequence_features = [
        Sequence('title', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('genres', SequenceEncoder(sep='|', min_cnt=1)),
        Sequence('clickedMovieIds',
                 SequenceEncoder(
                     sep='|', max_len=5,
                     word2idx=movie_word2idx, idx2word=movie_idx2word),
                 embedding_name='movieId'),
        Sequence('noClickedMovieIds',
                 SequenceEncoder(
                     sep='|', max_len=5,
                     word2idx=movie_word2idx, idx2word=movie_idx2word),
                 embedding_name='movieId')]

    attention_groups = [
        AttentionGroup(
            name='group1',
            pairs=[{'ad': 'movieId',
                    'pos_hist': 'clickedMovieIds',
                    'neg_hist': 'noClickedMovieIds'}],
            hidden_layers=[8, 4])]

    features = Features(
        number_features=number_features,
        category_features=category_features,
        sequence_features=sequence_features)

    dataloader, _ = prepare_dataloader(features)

    return dataloader, features, attention_groups


def test_gru_augru_neg_with_sharing_emb():
    dataloader, features, attention_groups = (
        create_test_data_with_sharing_emb())

    attention_groups[0].gru_type = 'AUGRU'

    model = DIEN(
        features, attention_groups=attention_groups,
        use_negsampling=True,
        num_classes=2, embedding_size=4, hidden_layers=(16, 8),
        final_activation='sigmoid', dropout=0.3)

    model(next(iter(dataloader)))



"""
Class for sequence features
"""

# Authors: Hongwei Zhang
# License: MIT


from .base import BaseFeature


class Sequence(BaseFeature):
    """
    Class for sequence features

    Parameters
    ----------
    name : str
        Name of this feature.

    column_flow : ColumnFlow
        ColumnFlow to transform this feature.

    embedding_name: str
        Embedding name for reference. Give same embedding name to features that
        share same embedding layer.

    embedding_size: int
        Dimension of embedding layer.

    vocab_size: int
        Provide vocab_size if this feature do not need to be pre-processed.
        vocab_size is only be used when column_flow is None.

    max_length: int
        max_length is only be used when column_flow is None.

    Attributes
    ----------
    name : str
        Name of this feature.

    column_flow : ColumnFlow
        ColumnFlow to transform this feature.

    embedding_name: str
        Embedding name for reference. Give same embedding name to features that
        share same embedding layer.

    embedding_size: int
        Dimension of embedding layer.
    """
    def __init__(self, name, column_flow,
                 embedding_name=None, embedding_size=None,
                 vocab_size=None, max_length=None):
        super().__init__(name=name, column_flow=column_flow)
        self.embedding_name = embedding_name if embedding_name else name
        self.embedding_size = embedding_size
        self._vocab_size = vocab_size
        self._max_length = max_length

    def dimension(self):
        """The dimension (vocab size) of sequence feature is the dimension
        of last transformer in ColumnFlow.
        """
        if self.column_flow is not None:
            return self.column_flow.transformers[-1].dimension()
        else:
            if self._vocab_size:
                return self._vocab_size
            else:
                raise RuntimeError(
                    "If param column_flow is not given, "
                    "vocab_size must be given")

    def max_length(self):
        """The max length of sequence feature is the max length
        of last transformer in ColumnFlow.
        """
        if self.column_flow is not None:
            return self.column_flow.transformers[-1].max_length()
        else:
            if self._max_length:
                return self._max_length
            else:
                raise RuntimeError(
                    "If param column_flow is not given, "
                    "max_length must be given")



"""
Multilayer perceptron torch module.
"""

# Authors: Hongwei Zhang
# License: MIT


from collections import OrderedDict

import torch.nn as nn


class MLP(nn.Module):
    """Multilayer perceptron torch module.

    Parameters
    ----------
    input_size : int
        Size of input.

    hidden_layers : iterable
        Hidden layer sizes.

    dropout : float
        Dropout rate.

    activation : str
        Name of activation function. ReLU, PReLU and Sigmoid are supported.
    """
    def __init__(self, input_size, hidden_layers,
                 dropout=0.0, batchnorm=True, activation='relu'):
        super(MLP, self).__init__()
        modules = OrderedDict()

        previous_size = input_size
        for index, hidden_layer in enumerate(hidden_layers):
            modules[f"dense{index}"] = nn.Linear(previous_size, hidden_layer)
            if batchnorm:
                modules[f"batchnorm{index}"] = nn.BatchNorm1d(hidden_layer)
            if activation:
                if activation.lower() == 'relu':
                    modules[f"activation{index}"] = nn.ReLU()
                elif activation.lower() == 'prelu':
                    modules[f"activation{index}"] = nn.PReLU()
                elif activation.lower() == 'sigmoid':
                    modules[f"activation{index}"] = nn.Sigmoid()
                else:
                    raise NotImplementedError(f"{activation} is not supported")
            if dropout:
                modules[f"dropout{index}"] = nn.Dropout(dropout)
            previous_size = hidden_layer
        self._sequential = nn.Sequential(modules)

    def forward(self, input):
        return self._sequential(input)



#! /usr/bin/env python


from setuptools import setup, find_packages
import prediction_flow


with open('README.md') as f:
    LONG_DESCRIPTION = f.read()

with open('requirements.txt') as f:
    INSTALL_REQUIRES = f.read().splitlines()

DISTNAME = 'prediction-flow'
DESCRIPTION = 'Deep-Learning based CTR models implemented by PyTorch'
MAINTAINER = 'Hongwei Zhang'
MAINTAINER_EMAIL = 'hw_zhang@outlook.com'
URL = 'https://github.com/GitHub-HongweiZhang/prediction-flow'
LICENSE = 'MIT'
VERSION = prediction_flow.__version__


def setup_package():
    setup(
        name=DISTNAME,
        packages=find_packages(),
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        description=DESCRIPTION,
        url=URL,
        version=VERSION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/markdown",
        python_requires='>=3.6',
        include_package_data=True,
        install_requires=INSTALL_REQUIRES,
        classifiers=(
            'Development Status :: 3 - Alpha',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Intended Audience :: Science/Research',
            'Intended Audience :: Developers',
            'Intended Audience :: Education',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Topic :: Software Development',
            'Topic :: Scientific/Engineering',
        ),
        license=LICENSE,
        keywords=[
            'torch', 'ctr prediction', 'deep learning',
            'deepfm', 'din', 'dnn', 'deep neural network']
    )


if __name__ == '__main__':
    setup_package()



import itertools
from collections import OrderedDict

import torch.nn as nn


class EmbeddingMixin:
    def build_embeddings(
            self, default_embedding_size, fixed_embedding_size=False):
        embeddings = OrderedDict()
        embedding_sizes = OrderedDict()

        for feature in itertools.chain(
                self.features.category_features,
                self.features.sequence_features):
            if feature.embedding_name not in embeddings:
                embedding_size = default_embedding_size
                if not fixed_embedding_size:
                    embedding_size = (feature.embedding_size
                                      if feature.embedding_size
                                      else default_embedding_size)

                embeddings[feature.embedding_name] = nn.Embedding(
                    feature.dimension(), embedding_size, padding_idx=0)
                embedding_sizes[feature.embedding_name] = embedding_size
                self.add_module(
                    f"embedding:{feature.embedding_name}",
                    embeddings[feature.embedding_name])

            if feature.name != feature.embedding_name:
                embeddings[feature.name] = embeddings[feature.embedding_name]
                embedding_sizes[feature.name] = (
                    embedding_sizes[feature.embedding_name])
                if feature.embedding_size and (
                        feature.embedding_size !=
                        embedding_sizes[feature.name]):
                    raise RuntimeWarning(
                        f"embedding_size of {feature.name} should be "
                        f"the same with {feature.embedding_name}")

        return (embeddings, embedding_sizes)



from prediction_flow.transformers.column import CategoryEncoder


def test_str_inputs():
    category_encoder = CategoryEncoder(min_cnt=1)

    input_terms = ['this', 'is', 'a', 'simple', 'test']

    category_encoder.fit(input_terms)

    transformed = category_encoder.transform(input_terms)

    assert set(transformed) == {1, 2, 3, 4, 5}
    assert category_encoder.dimension() == 7


def test_int_inputs():
    category_encoder = CategoryEncoder(min_cnt=1)

    input_terms = [345, 3434, 23, 88, 4]

    category_encoder.fit(input_terms)

    transformed = category_encoder.transform(input_terms)

    assert set(transformed) == {1, 2, 3, 4, 5}
    assert category_encoder.dimension() == 7


def test_unseen_inputs():
    category_encoder = CategoryEncoder(min_cnt=1)

    input_terms = [345, 3434, 23, 88, 4]

    category_encoder.fit(input_terms)

    transformed = category_encoder.transform([345, 5343])

    assert set(transformed) == {4, 6}



"""
Pooling layers.
"""

# Authors: Hongwei Zhang
# License: MIT


import torch
import torch.nn as nn


class MaxPooling(nn.Module):
    """Max Pooling.

    Parameters
    ----------
    dim : int
        The dimension to do pooling.

    Attributes
    ----------
    dim : int
        The dimension to do pooling.
    """
    def __init__(self, dim):
        super(MaxPooling, self).__init__()
        self.dim = dim

    def forward(self, input):
        return torch.max(input, self.dim)[0]


class SumPooling(nn.Module):
    """Sum Pooling.

    Parameters
    ----------
    dim : int
        The dimension to do pooling.

    Attributes
    ----------
    dim : int
        The dimension to do pooling.
    """
    def __init__(self, dim):
        super(SumPooling, self).__init__()
        self.dim = dim

    def forward(self, input):
        return torch.sum(input, self.dim)



from .number_feature import Number
from .category_feature import Category
from .sequence_feature import Sequence
from .features import Features


__all__ = [
    'Number',
    'Category',
    'Sequence',
    'Features'
]
